"""

gadget
======

Implements classes and functions for handling gadget files; you rarely
need to access this module directly as it will be invoked
automatically via pynbody.load.

"""


import configparser
import copy
import errno
import itertools
import os
import pathlib
import struct
import sys
import warnings

import numpy as np

from .. import array, config_parser, family, units
from . import SimSnap, namemapper

# This is set here and not in a config file because too many things break
# if it is not 6


N_TYPE = 6

_type_map = {}

for name, gtypes in config_parser.items('gadget-type-mapping'):
    try:
        gtypes = np.array([int(q) for q in gtypes.split(",")])
        if (gtypes >= N_TYPE).any() or (gtypes < 0).any():
            raise ValueError(
                "Type specified for family " + name + " is out of bounds (" + gtypes + ").")
        _type_map[family.get_family(name)] = gtypes
    except configparser.NoOptionError:
        pass

_name_map, _rev_name_map = namemapper.setup_name_maps(
    'gadget-name-mapping', gadget_blocks=True)
_translate_array_name = namemapper.name_map_function(_name_map, _rev_name_map)


def _to_raw(s):
    if isinstance(s, str) and sys.version_info[0] > 2:
        return s.encode('utf-8')
    else:
        return s


def gadget_type(fam):
    if isinstance(fam, list):
        l = []
        for sf in fam:
            l.extend(gadget_type(sf))
        return l
    elif fam is None:
        return list(np.arange(0, N_TYPE))
    else:
        return _type_map[fam]


class GadgetBlock:

    """Class to describe each block.
    Each block has a start, a length, and a length-per-particle"""

    def __init__(self, start=0, length=0, partlen=0, dtype=np.float32, p_types=np.zeros(N_TYPE, bool)):
        # Start of block in file
        self.start = start
        # Length of block in file
        self.length = length
        # Bytes per particle in file
        self.partlen = partlen
        # Data type of block
        self.data_type = dtype
        # Types of particle this block contains
        self.p_types = p_types


def _output_order_gadget(all_keys):

    out = []
    out_dregs = copy.copy(all_keys)
    for X in map(str.strip, config_parser.get('gadget-default-output', 'field-ordering').split(',')):
        if X in out_dregs:
            del out_dregs[out_dregs.index(X)]
            out.append(X)

    return out + out_dregs


def _construct_gadget_header(data, endian='='):
    """Create a GadgetHeader from a byte range read from a file."""
    npart = np.zeros(N_TYPE, dtype=np.uint32)
    mass = np.zeros(N_TYPE)
    time = 0.
    redshift = 0.
    npartTotal = np.zeros(N_TYPE, dtype=np.int32)
    num_files = 0
    BoxSize = 0.
    Omega0 = 0.
    OmegaLambda = 0.
    HubbleParam = 0.
    NallHW = np.zeros(N_TYPE, dtype=np.int32)
    if data == '':
        return
    fmt = endian + "IIIIIIddddddddiiIIIIIIiiddddiiIIIIIIiiif48s"
    if struct.calcsize(fmt) != 256:
        raise Exception(
            "There is a bug in gadget.py; the header format string is not 256 bytes")
    (npart[0], npart[1], npart[2], npart[3], npart[4], npart[5],
     mass[0], mass[1], mass[2], mass[3], mass[4], mass[5],
     time, redshift,  flag_sfr, flag_feedback,
     npartTotal[0], npartTotal[1], npartTotal[
     2], npartTotal[3], npartTotal[4], npartTotal[5],
     flag_cooling, num_files, BoxSize, Omega0, OmegaLambda, HubbleParam, flag_stellarage, flag_metals,
     NallHW[0], NallHW[1], NallHW[2], NallHW[3], NallHW[4], NallHW[5],
     flag_entropy_instead_u, flag_doubleprecision, flag_ic_info, lpt_scalingfactor, fill) = struct.unpack(fmt, data)

    header = GadgetHeader(npart, mass, time, redshift,
                          BoxSize, Omega0, OmegaLambda, HubbleParam, num_files)
    header.flag_sfr = flag_sfr
    header.flag_feedback = flag_feedback
    header.npartTotal = npartTotal
    header.flag_cooling = flag_cooling
    header.flag_stellarage = flag_stellarage
    header.flag_metals = flag_metals
    header.NallHW = NallHW
    header.flag_entropy_instead_u = flag_entropy_instead_u
    header.flag_doubleprecision = flag_doubleprecision
    header.flag_ic_info = flag_ic_info
    header.lpt_scalingfactor = lpt_scalingfactor
    header.endian = endian

    return header


class GadgetHeader:

    """Describes the header of gadget class files; this is all our metadata, so we are going to store it inline"""

    def __init__(self, npart, mass, time, redshift, BoxSize, Omega0, OmegaLambda, HubbleParam, num_files=1):
        "Construct a header from values, instead of a datastring."""
        assert(len(mass) == 6)
        assert(len(npart) == 6)
        # Mass of each particle type in this file. If zero,
        # particle mass stored in snapshot.
        self.mass = mass
        # Time of snapshot
        self.time = time
        # Redshift of snapshot
        self.redshift = redshift
        # Boolean to test the presence of star formation
        self.flag_sfr = False
        # Boolean to test the presence of feedback
        self.flag_feedback = False
        # Boolean to test the presence of cooling
        self.flag_cooling = False
        # Number of files expected in this snapshot
        self.num_files = num_files
        # Box size of the simulation
        self.BoxSize = BoxSize
        # Omega_Matter. Note this is Omega_DM + Omega_Baryons
        self.Omega0 = Omega0
        # Dark energy density
        self.OmegaLambda = OmegaLambda
        # Hubble parameter, in units where it is around 70.
        self.HubbleParam = HubbleParam
        # Boolean to test whether stars have an age
        self.flag_stellarage = False
        # Boolean to test the presence of metals
        self.flag_metals = False
        # flags that IC-file contains entropy instead of u
        self.flag_entropy_instead_u = False
        # flags that snapshot contains double-precision instead of single
        # precision
        self.flag_doubleprecision = False
        self.flag_ic_info = False
        # flag to inform whether IC files are generated with Zeldovich approximation,
        # or whether they contain 2nd order lagrangian perturbation theory ICs.
        #    FLAG_ZELDOVICH_ICS     (1)   - IC file based on Zeldovich
        #    FLAG_SECOND_ORDER_ICS  (2)   - Special IC-file containing 2lpt masses
        #    FLAG_EVOLVED_ZELDOVICH (3)   - snapshot evolved from Zeldovich ICs
        #    FLAG_EVOLVED_2LPT      (4)   - snapshot evolved from 2lpt ICs
        #    FLAG_NORMALICS_2LPT    (5)   - standard gadget file format with 2lpt ICs
        # All other values, including 0 are interpreted as "don't know" for
        # backwards compatability.
        self.lpt_scalingfactor = 0.    # scaling factor for 2lpt initial conditions
        self.endian = ""
        # Number of particles
        self.npart = np.array(npart, dtype=np.uint32)
        if (npart < 2 ** 31).all():
            # First 32-bits of total number of particles in the simulation
            self.npartTotal = np.array(npart, dtype=np.int32)
            # Long word of the total number of particles in the simulation.
            # At least one version of N-GenICs sets this to something entirely
            # different.
            self.NallHW = np.zeros(N_TYPE, dtype=np.int32)
        else:
            self.header.NallHW = np.array(npart // 2 ** 32, dtype=np.int32)
            self.header.npartTotal = np.array(
                npart - 2 ** 32 * self.header.NallHW, dtype=np.int32)

    def serialize(self):
        """This takes the header structure and returns it as a packed string"""
        fmt = self.endian + "IIIIIIddddddddiiIIIIIIiiddddiiIIIIIIiiif"
        # Do not attempt to include padding in the serialised data; the most common use of serialise
        # is to write to a file and we don't want to overwrite extra data that
        # might be present
        if struct.calcsize(fmt) != 256 - 48:
            raise Exception(
                "There is a bug in gadget.py; the header format string is not 256 bytes")
        # WARNING: On at least python 2.6.3 and numpy 1.3.0 on windows, castless code fails with:
        # SystemError: ..\Objects\longobject.c:336: bad argument to internal function
        # This is because self.npart, etc, has type np.uint32 and not int.
        # This is I think a problem with python/numpy, but cast things to ints
        # until I can determine how widespread it is.
        data = struct.pack(
            fmt, int(self.npart[0]), int(self.npart[1]), int(self.npart[
                                                             2]), int(
                                                                 self.npart[3]), int(
                                                                     self.npart[4]), int(
                                                                         self.npart[
                                                                             5]),
            self.mass[0], self.mass[1], self.mass[
                2], self.mass[3], self.mass[4], self.mass[5],
            self.time, self.redshift,  self.flag_sfr, self.flag_feedback,
            int(self.npartTotal[0]), int(self.npartTotal[1]), int(self.npartTotal[
                                                                  2]), int(
                                                                      self.npartTotal[3]), int(
                                                                          self.npartTotal[4]), int(
                                                                              self.npartTotal[
                                                                                  5]),
            self.flag_cooling, self.num_files, self.BoxSize, self.Omega0, self.OmegaLambda, self.HubbleParam, self.flag_stellarage, self.flag_metals,
            int(self.NallHW[0]), int(self.NallHW[1]), int(self.NallHW[
                                                          2]), int(
                                                              self.NallHW[3]), int(
                                                                  self.NallHW[4]), int(
                                                                      self.NallHW[
                                                                          5]),
            self.flag_entropy_instead_u, self.flag_doubleprecision, self.flag_ic_info, self.lpt_scalingfactor)
        return data


class GadgetFile:

    """Gadget file management class. Users should access gadget files through
    :class:`~pynbody.snapshot.gadget.GadgetSnap`."""

    def __init__(self, filename):
        self._filename = filename
        self.blocks = {}
        self.endian = ''
        self.format2 = True
        t_part = 0
        with open(filename, "rb") as fd:
            self.check_format(fd)
            # If format 1, load the block definitions.
            if not self.format2:
                self.block_names = config_parser.get(
                    'gadget-1-blocks', "blocks").split(",")
                self.block_names = [q.upper().ljust(4) for q in self.block_names]
                if sys.version_info[0] > 2:
                    self.block_names = [str.encode(x, 'utf-8') for x in self.block_names]
                # This is a counter for the fallback
                self.extra = 0
            while True:
                block = GadgetBlock()
                (name, block.length) = self.read_block_head(fd)
                if block.length == 0:
                    break
                # Do special things for the HEAD block
                if name[0:4] == b"HEAD":
                    if block.length != 256:
                        raise OSError("Mis-sized HEAD block in " + filename)
                    self.header = fd.read(256)
                    if len(self.header) != 256:
                        raise OSError("Could not read HEAD block in " + filename)
                    self.header = _construct_gadget_header(
                        self.header, self.endian)
                    record_size = self.read_block_foot(fd)
                    if record_size != 256:
                        raise OSError("Bad record size for HEAD in " + filename)
                    t_part = self.header.npart.sum()
                    if  ((not self.format2) and
                        ((self.header.npart != 0) * (self.header.mass == 0)).sum()==0):
                        # The "Spec" says that if all the existing particle masses
                        # are in the header, we shouldn't have a MASS block
                        self.block_names.remove(b"MASS")
                    continue
                # Set the partlen, using our amazing heuristics
                success = False
                if name[0:4] == b"POS " or name[0:4] == b"VEL ":
                    if block.length == t_part * 24:
                        block.partlen = 24
                        block.data_type = np.float64
                    else:
                        block.partlen = 12
                        block.data_type = np.float32
                    block.p_types = self.header.npart != 0
                    success = True
                elif name[0:4] == b"ID  ":
                    # Heuristic for long (64-bit) IDs
                    if block.length == t_part * 4:
                        block.partlen = 4
                        block.data_type = np.int32
                    else:
                        block.partlen = 8
                        block.data_type = np.int64
                    block.p_types = self.header.npart != 0
                    success = True

                block.start = fd.tell()
                # Check for the case where the record size overflows an int.
                # If this is true, we can't get record size from the length and we just have to guess
                # At least the record sizes at either end should be consistently wrong.
                # Better hope this only happens for blocks where all particles are
                # present.
                extra_len = int(t_part) * block.partlen
                if extra_len >= 2 ** 32:
                    fd.seek(extra_len, 1)
                else:
                    fd.seek(block.length, 1)
                record_size = self.read_block_foot(fd)
                if record_size != block.length:
                    raise OSError("Corrupt record in " +
                                filename + " footer for block " + name + "dtype" + str(block.data_type))
                if extra_len >= 2 ** 32:
                    block.length = extra_len

                if not success:
                    # Figure out what particles are here and what types
                    # they have. This also is a heuristic, which assumes
                    # that blocks are either fully present or not for a
                    # given particle. It also has to try all
                    # possibilities of dimensions of array and data type.
                    for dim, tp in (1, np.float32), (1, np.float64), (3, np.float32), (3, np.float64), (11, np.float32):
                        try:
                            block.data_type = tp
                            block.partlen = np.dtype(tp).itemsize * dim
                            block.p_types = self.get_block_types(
                                block, self.header.npart)
                            success = True
                            break
                        except ValueError:
                            continue

                if not success:
                    warnings.warn("Encountered a gadget block %r which could not be interpreted - is it a strange length or data type (length=%d)?" %
                                (name, block.length), RuntimeWarning)
                else:
                    self.blocks[name[0:4]] = block

        # Make a mass block if one isn't found.
        if b'MASS' not in self.blocks:
            block = GadgetBlock()
            block.length = 0
            block.start = 0

            # Mass should be the same type as POS (see issue #321)
            block.data_type = self.blocks[b'POS '].data_type
            block.partlen = np.dtype(block.data_type).itemsize
            self.blocks[b'MASS'] = block

    def get_block_types(self, block, npart):
        """ Set up the particle types in the block, with a heuristic,
        which assumes that blocks are either fully present or not for a given particle type"""

        totalnpart = npart.astype(np.int64).sum()
        if block.length == totalnpart * block.partlen:
            p_types = np.ones(N_TYPE, bool)
            return p_types
        p_types = np.zeros(N_TYPE, bool)
        for blocknpart in [1, 2, 3, 4, 5]:
            # iterate of differeent possible combinations of particles in the bloc
            # we stop when we can we match the length of the block
            for perm in itertools.permutations(list(range(0, N_TYPE)), blocknpart):
                # the 64-bit calculation is important here
                if block.length == (npart[list(perm)]).astype(np.int64).sum() * block.partlen:
                    p_types[list(perm)] = True
                    return p_types
        raise ValueError("Could not determine particle types for block")

    def check_format(self, fd):
        """This function reads the first character of a file and, depending on its value, determines
        whether we have a format 1 or 2 file, and whether the endianness is swapped. For the endianness,
        it then determines the correct byteorder string to pass to struct.unpack. There is not string
        for 'not native', so this is more complex than it needs to be"""
        fd.seek(0, 0)
        (r,) = struct.unpack('=I', fd.read(4))
        if r == 8:
            self.endian = '='
            self.format2 = True
        elif r == 134217728:
            if sys.byteorder == 'little':
                self.endian = '>'
            else:
                self.endian = '<'
            self.format2 = True
        elif r == 65536:
            if sys.byteorder == 'little':
                self.endian = '>'
            else:
                self.endian = '<'
            self.format2 = False
        elif r == 256:
            self.endian = '='
            self.format2 = False
        else:
            raise OSError("File corrupt. First integer is: " + str(r))
        fd.seek(0, 0)
        return

    def read_block_foot(self, fd):
        """Unpacks the block footer, into a single integer"""
        record_size = fd.read(4)
        if len(record_size) != 4:
            raise OSError("Could not read block footer")
        (record_size,) = struct.unpack(self.endian + 'I', record_size)
        return record_size

    def read_block_head(self, fd):
        """Read the Gadget 2 "block header" record, ie, 8 name, length, 8.
           Takes an open file and returns a (name, length) tuple """
        if self.format2:
            head = fd.read(5 * 4)
            # If we have run out of file, we don't want an exception,
            # we just want a zero length empty block
            if len(head) != 5 * 4:
                return ("    ", 0)
            head = struct.unpack(self.endian + 'I4sIII', head)
            if head[0] != 8 or head[3] != 8 or head[4] != head[2] - 8:
                raise OSError(
                    "Corrupt header record. Possibly incorrect file format")
            # Don't include the two "record_size" indicators in the total
            # length count
            return (head[1], head[2] - 8)
        else:
            record_size = fd.read(4)
            if len(record_size) != 4:
                return ("    ", 0)
            (record_size,) = struct.unpack(self.endian + 'I', record_size)
            try:
                name = self.block_names[0]
                self.block_names = self.block_names[1:]
            except IndexError:
                if self.extra == 0:
                    warnings.warn(
                        "Run out of block names in the config file. Using fallbacks: UNK*", RuntimeWarning)
                name = _to_raw("UNK" + str(self.extra))
                self.extra += 1
            return (name, record_size)

    def get_block(self, name, p_type, p_toread):
        """Get a particle range from this file, starting at p_start,
        and reading a maximum of p_toread particles"""
        name = _to_raw(name)

        p_read = 0
        cur_block = self.blocks[name]
        parts = self.get_block_parts(name, p_type)
        p_start = self.get_start_part(name, p_type)
        if p_toread > parts:
            p_toread = parts
        with open(self._filename, 'rb') as fd:
            fd.seek(cur_block.start + int(cur_block.partlen * p_start), 0)
            # This is just so that we can get a size for the type
            dt = np.dtype(cur_block.data_type)
            n_type = p_toread * cur_block.partlen // dt.itemsize
            data = np.fromfile(
                fd, dtype=cur_block.data_type, count=n_type, sep='')

        if self.endian != '=':
            data = data.byteswap(True)
        return (p_toread, data)

    def get_block_parts(self, name, p_type):
        """Get the number of particles present in a block in this file"""
        if name not in self.blocks:
            return 0
        cur_block = self.blocks[name]
        if p_type == -1:
            return cur_block.length // cur_block.partlen
        else:
            return self.header.npart[p_type] * cur_block.p_types[p_type]

    def get_start_part(self, name, p_type):
        """Find particle to skip to before starting, if reading particular type"""
        if p_type == -1:
            return 0
        else:
            if name not in self.blocks:
                return 0
            cur_block = self.blocks[name]
            return (cur_block.p_types * self.header.npart)[0:p_type].sum().astype(int)

    def get_block_dims(self, name):
        """Get the dimensionality of the block, eg, 3 for POS, 1 for most other things"""
        if name not in self.blocks:
            return 0
        cur_block = self.blocks[name]
        dt = np.dtype(cur_block.data_type)
        return cur_block.partlen // dt.itemsize

    # The following functions are for writing blocks back to the file
    def write_block(self, name, p_type, big_data, filename=None):
        """Write a full block of data in this file. Any particle type can be written. If the particle type is not present in this file,
        an exception KeyError is thrown. If there are too many particles, ValueError is thrown.
        big_data contains a reference to the data to be written. Type -1 is all types"""
        try:
            cur_block = self.blocks[name]
        except KeyError:
            raise KeyError("Block " + name + " not in file " + self._filename)

        if not cur_block.p_types[p_type]:
            return

        parts = self.get_block_parts(name, p_type)
        p_start = self.get_start_part(name, p_type)
        MinType = np.ravel(np.where(cur_block.p_types * self.header.npart))[0]
        MaxType = np.ravel(np.where(cur_block.p_types * self.header.npart))[-1]
        # Have we been given the right number of particles?
        if np.size(big_data) > parts * self.get_block_dims(name):
            raise ValueError("Space for " + str(parts) + " particles of type " + str(
                p_type) + " in file " + self._filename + ", " + str(np.shape(big_data)[0]) + " requested.")
        # Do we have the right type?
        dt = np.dtype(cur_block.data_type)
        bt = big_data.dtype
        if bt.kind != dt.kind:
            raise ValueError("Data of incorrect type passed to write_block")
        # Open the file
        fn = self._filename if filename is None else filename

        with open(fn, 'r+b') as fd:
            # Seek to the start of the block
            fd.seek(cur_block.start + cur_block.partlen * p_start, 0)
            # Add the block header if we are at the start of a block
            if p_type == MinType or p_type < 0:
                data = self.write_block_header(name, cur_block.length)
                # Better seek back a bit first.
                fd.seek(-len(data), 1)
                fd.write(data)

            if self.endian != '=':
                big_data = big_data.byteswap(False)

            # Actually write the data
            # Make sure to ravel it, otherwise the wrong amount will be written,
            # because it will also write nulls every time the first array dimension
            # changes.
            d = np.ravel(big_data.astype(dt)).tobytes()
            fd.write(d)
            if p_type == MaxType or p_type < 0:
                data = self.write_block_footer(name, cur_block.length)
                fd.write(data)

    def add_file_block(self, name, blocksize, partlen=4, dtype=np.float32, p_types=None):
        """Add a block to the block table at the end of the file. Do not actually write anything"""
        name = _to_raw(name)

        if name in self.blocks:
            raise KeyError(
                "Block " + name + " already present in file. Not adding")

        # Get last block
        lb = max(list(self.blocks.values()), key=lambda val: val.start)

        if np.issubdtype(dtype, np.floating):
            dtype = np.float32  # coerce to single precision

        # Make new block
        block = GadgetBlock(length=blocksize, partlen=partlen, dtype=dtype)
        block.start = lb.start + lb.length + 6 * \
            4  # For the block header, and footer of the previous block
        if p_types is None:
            block.p_types = np.ones(N_TYPE, bool)
        else:
            block.p_types = p_types
        self.blocks[name] = block

    def write_block_header(self, name, blocksize):
        """Create a string for a Gadget-style block header, but do not actually write it, for atomicity."""
        if self.format2:
            # This is the block header record, which we want for format two
            # files only
            blkheadsize = 4 + 4 * 1
            # 1 int and 4 chars
            nextblock = blocksize + 2 * 4
            # Relative location of next block; the extra 2 uints are for storing the headers.
            # Write format 2 header header
            head = struct.pack(
                self.endian + 'I4sII', blkheadsize, name, nextblock, blkheadsize)
        # Also write the record size, which we want for all files*/
        head += self.write_block_footer(name, blocksize)
        return head

    def write_block_footer(self, name, blocksize):
        """(Re) write a Gadget-style block footer."""
        return struct.pack(self.endian + 'I', blocksize)

    def write_header(self, head_in, filename=None):
        """Write a file header. Overwrites npart in the argument with the npart of the file, so a consistent file is always written."""
        # Construct new header with the passed header and overwrite npart with the file header.
        # This has ref. semantics so use copy
        head = copy.deepcopy(head_in)
        head.npart = np.array(self.header.npart)
        data = self.write_block_header(b"HEAD", 256)
        data += head.serialize()
        if filename is None:
            filename = self._filename

        try:
            fd = open(filename, "r+b")
        except OSError as error:
            # If we couldn't open it because it doesn't exist open it for
            # writing.
            (err, strerror) = error.args
            # If we couldn't open it because it doesn't exist open it for
            # writing.
            if err == errno.ENOENT:
                fd = open(filename, "w+b")
            # If we couldn't open it for any other reason, reraise exception
            else:
                raise OSError(err, strerror)

        try:
            fd.seek(0)  # Header always at start of file
            # Write header
            fd.write(data)
            # Seek 48 bytes forward, to skip the padding (which may contain extra
            # data)
            fd.seek(48, 1)
            data = self.write_block_footer(b"HEAD", 256)
            fd.write(data)
        finally:
            fd.close()


class GadgetWriteFile (GadgetFile):

    """Class for write-only snapshots, as when we are creating a new set of files from, eg, a TipsySnap.
        Should not be used directly. block_names is a list so we can specify an on-disc ordering."""

    def __init__(self, filename, npart, block_names, header, format2=True):
        self.header = header
        self._filename = filename
        self.endian = '='  # write with default endian of this system
        self.format2 = format2
        self.blocks = {}
        self.header.npart = np.array(npart)
        # Set up the positions
        header_size = 4
        if format2:
            header_size += 3 * 4 + 4
        footer_size = 4
        # First block is just past the header.
        cur_pos = 256 + header_size + footer_size
        for block in block_names:
            # Add block if present for some types
            if block.types.sum():
                b_part = npart * block.types
                b = GadgetBlock(
                    start=cur_pos + header_size, partlen=block.partlen,
                    length=block.partlen * b_part.sum(), dtype=block.dtype, p_types=block.types)
                cur_pos += b.length + header_size + footer_size
                self.blocks[_to_raw(block.name)] = b


class WriteBlock:

    """Internal structure for passing data around between file and snapshot"""

    def __init__(self, partlen=4, dtype=np.float32, types=np.zeros(N_TYPE, bool), name="    "):

        if np.issubdtype(dtype, np.floating):
            dtype = np.float32
        if np.issubdtype(dtype, np.signedinteger):
            dtype = np.int32

        self.partlen = partlen * np.dtype(dtype).itemsize
        self.dtype = dtype

        self.types = types
        self.name = name


class GadgetSnap(SimSnap):

    """Main class for reading Gadget-2 snapshots. The constructor makes a map of the locations
    of the blocks, which are then read by _load_array"""

    def __init__(self, filename: pathlib.Path, only_header=False, must_have_paramfile=False, ignore_cosmo=False):

        filename = str(filename)

        global config
        super().__init__()
        self._files = []
        self._filename = filename
        self._ignore_cosmo = ignore_cosmo

        # Check whether the file exists, and get the ".0" right
        if os.path.exists(filename):
            files = [filename]
        elif os.path.exists(filename + ".0"):
            filename = filename + ".0"
            files = None

        if filename[-2:] == ".0":
            self._filename = filename[:-2]
        # Read the first file and use it to get an idea of how many files we
        # are expecting.
        first_file = GadgetFile(filename)
        self._files.append(first_file)
        files_expected = self._files[0].header.num_files
        npart = np.array(self._files[0].header.npart)

        if files is None:
            # we want to load all files
            base_filename = filename[:-2]
            files = [base_filename + "." + str(i)
                     for i in range(files_expected)]

        for filename in files[1:]:
            tmp_file = GadgetFile(filename)
            if not self.check_headers(tmp_file.header, self._files[0].header):
                warnings.warn("file " + str(
                    i) + " is not part of this snapshot set!", RuntimeWarning)
                continue
            self._files.append(tmp_file)
            npart = npart + tmp_file.header.npart
        # Set up things from the parent class
        self._num_particles = npart.sum()
        # Set up global header
        self.header = copy.deepcopy(self._files[0].header)
        self.header.npart = npart
        # Check and fix npartTotal and NallHW if they are wrong.
        if npart is not self.header.npartTotal + 2 ** 32 * self.header.NallHW:
            self.header.NallHW = npart // 2 ** 32
            self.header.npartTotal = npart - 2 ** 32 * self.header.NallHW
            for f in self._files:
                f.header.npartTotal = self.header.npartTotal
                f.header.NallHW = self.header.NallHW

        self._family_slice = {}

        self._loadable_keys = set()
        self._family_keys = set()
        self._family_arrays = {}
        self._arrays = {}
        #self.properties = {}

        # Set up _family_slice
        current = 0
        for fam in _type_map:
            g_types = _type_map[fam]
            length = 0
            for f in self._files:
                length += sum(f.header.npart[x] for x in g_types)
            self._family_slice[fam] = slice(current, current + length)
            current += length

        # Set up _loadable_keys
        for f in self._files:
            self._loadable_keys = self._loadable_keys.union(
                set(f.blocks.keys()))

        # Add default mapping to unpadded lower case if not in config file.
        for nn in self._loadable_keys:
            if sys.version_info[0] == 2:
                mm = nn.lower().strip()
            else:
                mm = nn.lower().strip().decode('utf-8')
            if nn not in _rev_name_map:
                _rev_name_map[nn] = mm
            if mm not in _name_map:
                _name_map[mm] = nn

        # Use translated keys only
        self._loadable_keys = [_translate_array_name(
            x, reverse=True) for x in self._loadable_keys]
        # Set up block list, with attached families, as a caching mechanism
        self._block_list = self.get_block_list()

        self._decorate()

    def loadable_family_keys(self, fam=None):
        """Return list of arrays which are loadable for specific families,
        but not for all families."""
        warnings.warn(
            "loadable_family_keys functionality has now been incorporated into loadable_keys", warnings.DeprecationWarning)
        return self.loadable_keys(fam)

    def loadable_keys(self, fam=None):
        if fam is not None:
            return [x for x in self._loadable_keys if self._family_has_loadable_array(fam, x)]
        else:
            return [x for x in self._loadable_keys if self._family_has_loadable_array(None, x)]

    def _family_has_loadable_array(self, fam, name):
        """Returns True if the array can be loaded for the specified family.
        If fam is None, returns True if the array can be loaded for all families."""
        if name in self._block_list:
            if fam is not None:
                return fam in self._block_list[name]
            else:
                return set(self.families()) <= set(self._block_list[name])
        else:
            return False

    def get_block_list(self):
        """Get list of unique blocks in snapshot, with the types they refer to"""
        b_list = {}
        for f in self._files:
            for (n, b) in f.blocks.items():
                if n in b_list:
                    b_list[n] += b.p_types
                else:
                    b_list[n] = np.array(b.p_types, dtype=bool)
        # Special case mass. Note b_list has reference semantics.
        if b"MASS" in b_list:
            b_list[b"MASS"] += np.array(self.header.mass, dtype=bool)
        # Translate this array into families and external names
        out_list = {}
        for k, b in b_list.items():
            b_name = _translate_array_name(k, reverse=True)
            # Make this be only if there are actually particles of that type in
            # the snap
            b_types = [f for f in self.families() if b[np.intersect1d(
                gadget_type(f), np.ravel(np.where(self.header.npart != 0)))].all()]
            out_list[b_name] = b_types
        return out_list

    def get_block_parts(self, name, family):
        """Get the number of particles present in a block, of a given type"""
        total = 0
        for f in self._files:
            total += sum(f.get_block_parts(
                name, gfam) for gfam in gadget_type(family))
        # Special-case MASS
        if name == b"MASS":
            total += sum(self.header.npart[p] * np.array(self.header.mass[
                         p], dtype=bool) for p in gadget_type(family))
        return total

    def check_headers(self, head1, head2):
        """Check two headers for consistency"""
        if (head1.time != head2.time or head1.redshift != head2.redshift or
                head1.flag_sfr != head2.flag_sfr or
                head1.flag_feedback != head2.flag_feedback or
                head1.num_files != head2.num_files or
                head1.BoxSize != head2.BoxSize or
                head1.Omega0 != head2.Omega0 or
                head1.OmegaLambda != head2.OmegaLambda or
                head1.HubbleParam != head2.HubbleParam or
                head1.flag_stellarage != head2.flag_stellarage or
                head1.flag_metals != head2.flag_metals):
            return False
        # Check array quantities
        if (((head1.mass - head2.mass) > 1e-5 * head1.mass).any() or
                (head1.npartTotal != head2.npartTotal).any()):
            return False
        #  At least one version of N-GenICs writes a header file which
        #  ignores everything past flag_metals (!), leaving it uninitialised.
        #  Therefore, we can't check them.
        return True

    def _get_array_type(self, name):
        """Get the type for the array given in name"""
        g_name = _translate_array_name(name)
        return self._get_array_type_g(g_name)

    def _get_array_type_g(self, name):
        """Get the type for the array given in name"""
        return self._files[0].blocks[name].data_type

    def _get_array_dims(self, name):
        """Get the dimensions of an array; ie, is it 3d or 1d"""
        g_name = _translate_array_name(name)
        return self._files[0].get_block_dims(g_name)

    def _load_array(self, name, fam=None):
        """Read in data from a Gadget file.
        If fam is not None, loads only data for that particle family"""
        # g_name is the internal name
        g_name = _translate_array_name(name)

        if not self._family_has_loadable_array(fam, name):
            if fam is None and name in self._block_list:
                raise KeyError(
                    "Block " + name + " is not available for all families")
            else:
                raise OSError("No such array on disk")

        ndim = self._get_array_dims(name)

        if ndim == 1:
            dims = [self.get_block_parts(g_name, fam), ]
        else:
            dims = [self.get_block_parts(g_name, fam), ndim]

        if fam is not None:
            p_types = gadget_type(fam)
        else:
            p_types = gadget_type(self.families())

        # Get the data. Get one type at a time and then concatenate.
        # A possible optimisation is to special-case loading all particles.
        data = np.array([], dtype=self._get_array_type(name))
        for p in p_types:
            # Special-case mass
            if g_name == b"MASS" and self.header.mass[p] != 0.:
                data = np.append(data, self.header.mass[
                                 p] * np.ones(self.header.npart[p], dtype=data.dtype))
            else:
                data = np.append(data, self.__load_array(g_name, p))

        if fam is None:
            self[name] = data.reshape(dims, order='C').view(array.SimArray)
            self[name].set_default_units(quiet=True)
        else:
            self[fam][name] = data.reshape(
                dims, order='C').view(array.SimArray)
            self[fam][name].set_default_units(quiet=True)

    def __load_array(self, g_name, p_type):
       """Internal helper function for _load_array that takes a g_name and a gadget type,
       gets the data from each file and returns it as one long array."""
       data = np.array(np.zeros(self._get_array_dims(g_name) * self.header.npart[p_type]), dtype=self._get_array_type_g(g_name))
       # Get a type from each file
       ipos = 0
       for f in self._files:
           f_parts = f.get_block_parts(g_name, p_type)
           if f_parts == 0:
               continue
           (f_read, f_data) = f.get_block(g_name, p_type, f_parts)
           if f_read != f_parts:
               raise OSError("Read of " + f._filename + " asked for " + str(
                   f_parts) + " particles but got " + str(f_read))
           iread = self._get_array_dims(g_name)*f.header.npart[p_type]
           data[ipos:ipos + iread] = f_data
           ipos += iread
       return data

    @classmethod
    def _can_load(cls, f: pathlib.Path):
        """Check whether we can load the file as Gadget format by reading
        the first 4 bytes"""
        fname = f
        if not f.exists():
            if not f.with_suffix(".0").exists():
                return False
            fname = f.with_suffix(".0")

        with open(fname, "br") as fd:
            r, = struct.unpack('=I', fd.read(4))

        # First int32 is 8 for a Gadget 2 file, or 256 for Gadget 1, or the
        # byte swapped equivalent.
        if r in (8, 134217728, 65536, 256):
            return True
        else:
            return False

    @staticmethod
    def _write(self, filename=None):
        """Write an entire Gadget file (actually an entire set of snapshots)."""

        with self.lazy_derive_off:
            # If caller is not a GadgetSnap, construct the GadgetFiles,
            # so that format conversion works.
            all_keys = set(self.loadable_keys()).union(
                list(self.keys())).union(self.family_keys())
            all_keys = [
                k for k in all_keys if k not in ["x", "y", "z", "vx", "vy", "vz"]]
            # This code supports (limited) format conversions
            if self.__class__ is not GadgetSnap:
                # We need a filename if we are writing to a new type
                if filename is None:
                    raise Exception(
                        "Please specify a filename to write a new file.")
            # Splitting the files correctly is hard; the particles need to be reordered, and
                # we need to know which families correspond to which gadget types.
                # So don't do it.

                # Make sure the data fits into one files. The magic numbers are:
                # 12 - the largest block is likely to  be POS with 12 bytes per particle.
                # 2**31 is the largest size a gadget block can safely have
                if self.__len__() * 12. > 2 ** 31 - 1:
                    raise OSError(
                        "Data too large to fit into a single gadget file, and splitting not implemented. Cannot write.")
                # Make npart
                npart = np.zeros(N_TYPE, int)
                arr_name = (list(self.keys()) + self.loadable_keys())[0]
                for f in self.families():
                    # Note that if we have more than one type per family, we cannot
                    # determine which type each individual particle is, so
                    # assume they are all the first.
                    npart[np.min(gadget_type(f))] = len(self[f][arr_name])
                # Construct a header
                # npart, mass, time, redshift, BoxSize,Omega0, OmegaLambda,
                # HubbleParam, num_files=1
                gheader = GadgetHeader(
                    npart, np.zeros(N_TYPE, float), self.properties[
                        "a"], self.properties["z"],
                    self.properties["boxsize"].in_units(self[
                                                        'pos'].units, **self.conversion_context()),
                    self.properties["omegaM0"], self.properties["omegaL0"], self.properties["h"], 1)
                # Construct the block_names; each block_name needs partlen, data_type, and p_types,
                # as well as a name. Blocks will hit the disc in the order they are in all_keys.
                # First, make pos the first block and vel the second.

                # all_keys[all_keys.index("pos")]=all_keys[0]
                # all_keys[0] = "pos"
                # all_keys[all_keys.index("vel")]=all_keys[1]
                # all_keys[1] = "vel"
                all_keys = _output_order_gadget(all_keys)

                # No writing format 1 files.
                block_names = []
                for k in all_keys:
                    types = np.zeros(N_TYPE, bool)
                    for f in self.families():
                        try:
                            dtype = self[f][k].dtype
                            types[np.min(gadget_type(f))] += True
                            try:
                                partlen = np.shape(self[
                                                   f][k])[1]  # *dtype.itemsize
                            except IndexError:
                                partlen = 1  # dtype.itemsize
                        except KeyError:
                            pass
                    bb = WriteBlock(partlen, dtype=dtype, types=types, name=_translate_array_name(
                        k).upper().ljust(4)[0:4])
                    block_names.append(bb)
                # Create an output file
                out_file = GadgetWriteFile(
                    filename, npart, block_names, gheader)
                # Write the header
                out_file.write_header(gheader, filename)
                # Write all the arrays
                for x in all_keys:
                    g_name = _to_raw(
                        _translate_array_name(x).upper().ljust(4)[0:4])

                    for fam in self.families():
                        try:
                            data = self[fam][x]
                            gfam = np.min(gadget_type(fam))
                            out_file.write_block(
                                g_name, gfam, data, filename=filename)
                        except KeyError:
                            pass
                return

            # Write headers
            if filename is not None:
                if np.size(self._files) > 1:
                    for i in np.arange(0, np.size(self._files)):
                        ffile = filename + "." + str(i)
                        self._files[i].write_header(self.header, ffile)
                else:
                    self._files[0].write_header(self.header, filename)
            else:
                # Call write_header for every file.
                [f.write_header(self.header) for f in self._files]
            # Call _write_array for every array.
            for x in all_keys:
                GadgetSnap._write_array(self, x, filename=filename)

    @staticmethod
    def _write_array(self, array_name, fam=None, filename=None):
        """Write a data array back to a Gadget snapshot, splitting it across files."""
        write_fam = fam or self.families()

        # Make the name a four-character upper case name, possibly with
        # trailing spaces
        g_name = _to_raw(
            _translate_array_name(array_name).upper().ljust(4)[0:4])
        nfiles = np.size(self._files)
        # Find where each particle goes
        f_parts = [f.get_block_parts(g_name, -1) for f in self._files]
        # If there is no block corresponding to this name in the file,
        # add it (so we can write derived arrays).
        if np.sum(f_parts) == 0:
            # Get p_type
            p_types = np.zeros(N_TYPE, bool)
            npart = 0
            for fam in self.families():
                gfam = np.min(gadget_type(fam))
                # We get the particle types we want by trying to load all
                # particle types (from memory) and seeing which ones work
                p_types[gfam] = array_name in self[fam]
                if p_types[gfam]:
                    ashape = np.shape(self[fam][array_name])
                    # If the partlen is 1, append so the shape array has the
                    # right shape.
                    if np.size(ashape) < 2:
                        ashape = (ashape[0], 1)
                    npart += ashape[0]
            if p_types.sum():
                per_file = npart // nfiles
                for f in self._files[:-2]:
                    f.add_file_block(array_name, per_file, ashape[
                                     1], dtype=self[array_name].dtype, p_types=p_types)
                self._files[-1].add_file_block(
                    array_name, npart - (nfiles - 1) * per_file, ashape[1])

        # Write blocks on a family level, so that we don't have to worry about
        # the file-level re-ordering.
        for fam in write_fam:
            if self._family_has_loadable_array(fam, array_name):
                data = self[fam][array_name]
                s = 0
                for gfam in gadget_type(fam):
                    # Find where each particle goes
                    f_parts = [f.get_block_parts(
                        g_name, gfam) for f in self._files]
                    for i in np.arange(0, nfiles):
                        if f_parts[i]==0:
                            continue
                        # Set up filename
                        if filename is not None:
                            ffile = filename + "." + str(i)
                            if nfiles == 1:
                                ffile = filename
                        else:
                            ffile = None
                        # Special-case MASS.
                        if g_name == b"MASS" and self.header.mass[gfam] != 0.:
                            nmass = np.min(
                                data[s:(s + self.header.npart[gfam])])
                            # Warn if there are now different masses for this particle type,
                            # as this information cannot be represented in this
                            # snapshot.
                            if nmass != np.max(data[s:(s + self.header.npart[gfam])]):
                                warnings.warn("Cannot write variable masses for type " + str(
                                    gfam) + ", as masses are stored in the header.", RuntimeWarning)
                            elif self.header.mass[gfam] != nmass:
                                self.header.mass[gfam] = nmass
                                self._files[i].write_header(
                                    self.header, filename=ffile)
                        else:
                            # Write data
                            if np.issubdtype(data.dtype, np.floating):
                                data = np.asanyarray(data, dtype=np.float32)
                            self._files[i].write_block(g_name, gfam, data[
                                                       s:(s + f_parts[i])], filename=ffile)
                        s += f_parts[i]


def _header_suggests_cosmological(gadget_header):
    return (gadget_header.HubbleParam != 0.) and (gadget_header.Omega0 != 0.) and (gadget_header.BoxSize != 0.)

@GadgetSnap.decorator
def do_units(sim):
    # cosmo =
    # (sim._hdf_files['Parameters']['NumericalParameters'].attrs['ComovingIntegrationOn'])!=0

    vel_unit = config_parser.get('gadget-units', 'vel')
    dist_unit = config_parser.get('gadget-units', 'pos')
    mass_unit = config_parser.get('gadget-units', 'mass')

    vel_unit, dist_unit, mass_unit = (
        units.Unit(x) for x in (vel_unit, dist_unit, mass_unit))

    if sim._ignore_cosmo or not _header_suggests_cosmological(sim.header):
        # remove a and h dependences
        vel_unit = units.Unit(
            "km s^-1") * vel_unit.in_units("km s^-1", a=1, h=1)
        mass_unit = units.Unit("Msol") * mass_unit.in_units("Msol", a=1, h=1)
        dist_unit = units.Unit("kpc") * dist_unit.in_units("kpc", a=1, h=1)

    sim._file_units_system = [units.Unit("K"), vel_unit, dist_unit, mass_unit]
    sim._override_units_system()


@GadgetSnap.decorator
def do_properties(sim):
    h = sim.header

    if not(sim._ignore_cosmo) and _header_suggests_cosmological(h):
        from .. import analysis
        sim.properties['omegaM0'] = h.Omega0
        # sim.properties['omegaB0'] = ... This one is non-trivial to calculate
        sim.properties['omegaL0'] = h.OmegaLambda
        sim.properties['boxsize'] = sim.infer_original_units("kpc")*h.BoxSize
        sim.properties['z'] = h.redshift
        sim.properties['h'] = h.HubbleParam
        time_units = sim.infer_original_units("s")
        sim.properties['time'] = analysis.cosmology.age(sim,unit=time_units)*time_units
    else:
        sim.properties['time'] = sim.infer_original_units("s") * h.time



"""

generic
=======

Flexible and general plotting functions

"""

import numpy as np
import pylab as plt

import pynbody

from .. import config
from ..array import SimArray
from ..units import NoUnit


def qprof(sim, qty='metals', weights=None, q=(0.16, 0.5, 0.84),
          ax=False, ylabel=None, xlog=False, ylog=False, xlabel=None,
          facecolor='#BBBBBB', color='#BBBBBB', medcolor='black', filename=False):
    if ax:
        p = ax
        f = plt.gca()
    else:
        f, p = plt.subplots(1)

    qp = pynbody.analysis.profile.QuantileProfile(sim, q=q, weights=weights)
    p.fill_between(qp['rbins'].in_units('kpc'), qp[qty][:, 0], y2=qp[qty][:, 2],
                   facecolor=facecolor, color=color,
                   where=np.isfinite(qp[qty][:, 0]))
    p.plot(qp['rbins'].in_units('kpc'), qp[qty][:, 1], color=medcolor)
    if xlabel is None:
        p.set_xlabel('r [kpc]')
    else:
        p.set_xlabel(xlabel)
    if ylabel is not None:
        p.set_ylabel(ylabel)
    if xlog:
        p.semilogx()
    if ylog:
        p.semilogy()

    if filename:
        f.savefig(filename)


def hist2d(xo, yo, weights=None, mass=None, gridsize=(100, 100), nbins = None, make_plot = True, **kwargs):
    """
    Plot 2D histogram for arbitrary arrays that get passed in.

    **Input:**

       *x*: array

       *y*: array

    **Optional keywords:**

       *x_range*: list, array, or tuple
         size(x_range) must be 2. Specifies the X range.

       *y_range*: tuple
         size(y_range) must be 2. Specifies the Y range.

       *gridsize*: (int, int) (default (100,100))
         number of bins to use for the 2D histogram

       *nbins*: int
         number of bins for the histogram - if specified, gridsize is set to (nbins,nbins)

       *nlevels*: int
         number of levels to use for the contours

       *logscale*: boolean
         whether to use log or linear spaced contours

       *weights*: numpy array of same length as x and y
         if weights is passed, color corresponds to
         the mean value of weights in each cell

       *mass*: numpy array of masses same length as x andy
         must also have weights passed in.  If you just
         want to weight by mass, pass the masses to weights

       *colorbar*: boolean
         draw a colorbar

       *scalemin*: float
         minimum value to use for the color scale

       *scalemax*: float
         maximum value to use for the color scale
    """
    global config

    # process keywords
    x_range = kwargs.get('x_range', None)
    y_range = kwargs.get('y_range', None)
    xlogrange = kwargs.get('xlogrange', False)
    ylogrange = kwargs.get('ylogrange', False)
    ret_im = kwargs.get('ret_im', False)

    if y_range is not None:
        if len(y_range) != 2:
            raise RuntimeError("Range must be a length 2 list or array")
    else:
        if ylogrange:
            y_range = [np.log10(np.min(yo)), np.log10(np.max(yo))]
        else:
            y_range = [np.min(yo), np.max(yo)]
        kwargs['y_range'] = y_range

    if x_range is not None:
        if len(x_range) != 2:
            raise RuntimeError("Range must be a length 2 list or array")
    else:
        if xlogrange:
            x_range = [np.log10(np.min(xo)), np.log10(np.max(xo))]
        else:
            x_range = [np.min(xo), np.max(xo)]
        kwargs['x_range'] = x_range

    if (xlogrange):
        x = np.log10(xo)
    else:
        x = xo

    if (ylogrange):
        y = np.log10(yo)
    else:
        y = yo

    if nbins is not None:
        gridsize = (nbins, nbins)

    ind = np.where((x > x_range[0]) & (x < x_range[1]) &
                   (y > y_range[0]) & (y < y_range[1]))

    x = x[ind[0]]
    y = y[ind[0]]

    draw_contours = False
    if weights is not None and mass is not None:
        draw_contours = True
        weights = weights[ind[0]]
        mass = mass[ind[0]]

        # produce a mass-weighted histogram of average weight values at each
        # bin
        hist, ys, xs = np.histogram2d(
            y, x, weights=weights * mass, bins=gridsize, range=[y_range, x_range])
        hist_mass, ys, xs = np.histogram2d(
            y, x, weights=mass, bins=gridsize, range=[y_range, x_range])
        good = np.where(hist_mass > 0)
        hist[good] = hist[good] / hist_mass[good]

    else:
        if weights is not None:
            # produce a weighted histogram
            weights = weights[ind[0]]
        elif mass is not None:
            # produce a mass histogram
            weights = mass[ind[0]]

        hist, ys, xs = np.histogram2d(
            y, x, weights=weights, bins=gridsize, range=[y_range, x_range])

    try:
        hist = SimArray(hist, weights.units)
    except AttributeError:
        hist = SimArray(hist)

    try:
        xs = SimArray(.5 * (xs[:-1] + xs[1:]), x.units)
        ys = SimArray(.5 * (ys[:-1] + ys[1:]), y.units)
    except AttributeError:
        xs = .5 * (xs[:-1] + xs[1:])
        ys = .5 * (ys[:-1] + ys[1:])

    if ret_im:
        return make_contour_plot(hist, xs, ys, **kwargs)

    if make_plot:
        make_contour_plot(hist, xs, ys, **kwargs)
        if draw_contours:
            make_contour_plot(SimArray(density_mass, mass.units), xs, ys, filled=False,
                              clear=False, colorbar=False, colors='black', scalemin=nmin, nlevels=10)

    return hist, xs, ys


def gauss_kde(xo, yo, weights=None, mass=None, gridsize=(100, 100), nbins = None,
              make_plot = True, nmin = None, nmax = None, **kwargs):
    """
    Plot 2D gaussian kernel density estimate (KDE) given values at points (*x*, *y*).

    Behavior changes depending on which keywords are passed:

    If a *weights* array is supplied, produce a weighted KDE.

    If a *mass* array is supplied, a mass density is computed.

    If both *weights* and *mass* are supplied, a mass-averaged KDE of the weights is
    computed.

    By default, norm=False is passed to :func:`~pynbody.plot.util.fast_kde` meaning
    that the result returned *is not* normalized such that the integral over the area
    equals one.

Since this function produces a density estimate, the units of the
    output grid are different than for the output of
    :func:`~pynbody.plot.generic.hist2d`. To get to the same units,
    you must multiply by the size of the cells.


    **Input:**

       *xo*: array

       *yo*: array

    **Optional keywords:**

       *mass*: numpy array of same length as x and y
         particle masses to be used for weighting

       *weights*: numpy array of same length as x and y
         if weights is passed, color corresponds to
         the mean value of weights in each cell

       *nmin*: float (default None)
         if *weights* and *mass* are both specified, the mass-weighted
         contours are only drawn where the mass exceeds *nmin*.

       *gridsize*: (int, int) (default: 100,100)
         size of grid for computing the density estimate

       *nbins*: int
         number of bins for the histogram - if specified, gridsize is set to (nbins,nbins)

       *make_plot*: boolean (default: True)
         whether or not to produce a plot

    **Keywords passed to** :func:`~pynbody.plot.util.fast_kde`:

       *norm*: boolean (default: False)
         If False, the output is only corrected for the kernel. If True,
         the result is normalized such that the integral over the area
         yields 1.

       *nocorrelation*: (default: False) If True, the correlation
         between the x and y coords will be ignored when preforming
         the KDE.

    **Keywords passed to** :func:`~pynbody.plot.generic.make_contour_plot`:

       *x_range*: list, array, or tuple
         size(x_range) must be 2. Specifies the X range.

       *y_range*: tuple
         size(y_range) must be 2. Specifies the Y range.

       *nlevels*: int
         number of levels to use for the contours

       *logscale*: boolean
         whether to use log or linear spaced contours

       *colorbar*: boolean
         draw a colorbar

       *scalemin*: float
         minimum value to use for the color scale

       *scalemax*: float
         maximum value to use for the color scale
    """

    from .util import fast_kde

    global config

    # process keywords
    x_range = kwargs.get('x_range', None)
    y_range = kwargs.get('y_range', None)
    xlogrange = kwargs.get('xlogrange', False)
    ylogrange = kwargs.get('ylogrange', False)
    ret_im = kwargs.get('ret_im', False)

    if y_range is not None:
        if len(y_range) != 2:
            raise RuntimeError("Range must be a length 2 list or array")
    else:
        if ylogrange:
            y_range = [np.log10(np.min(yo)), np.log10(np.max(yo))]
        else:
            y_range = [np.min(yo), np.max(yo)]

    if x_range is not None:
        if len(x_range) != 2:
            raise RuntimeError("Range must be a length 2 list or array")
    else:
        if xlogrange:
            x_range = [np.log10(np.min(xo)), np.log10(np.max(xo))]
        else:
            x_range = [np.min(xo), np.max(xo)]

    if (xlogrange):
        x = np.log10(xo)
    else:
        x = xo
    if (ylogrange):
        y = np.log10(yo)
    else:
        y = yo

    if nbins is not None:
        gridsize = (nbins, nbins)

    ind = np.where((x > x_range[0]) & (x < x_range[1]) &
                   (y > y_range[0]) & (y < y_range[1]))

    x = x[ind[0]]
    y = y[ind[0]]

    try:
        xs = SimArray(
            np.linspace(x_range[0], x_range[1], gridsize[0] + 1), x.units)
    except AttributeError:
        xs = np.linspace(x_range[0], x_range[1], gridsize[0] + 1)
    xs = .5 * (xs[:-1] + xs[1:])
    try:
        ys = SimArray(
            np.linspace(y_range[0], y_range[1], gridsize[1] + 1), y.units)
    except AttributeError:
        ys = np.linspace(y_range[0], y_range[1], gridsize[1] + 1)
    ys = .5 * (ys[:-1] + ys[1:])

    extents = [x_range[0], x_range[1], y_range[0], y_range[1]]

    draw_contours = False
    if weights is not None and mass is not None:
        draw_contours = True
        weights = weights[ind[0]]
        mass = mass[ind[0]]

        # produce a mass-weighted gaussian KDE of average weight values at each
        # bin
        density = fast_kde(
            x, y, weights=weights * mass, gridsize=gridsize, extents=extents, **kwargs)
        density_mass = fast_kde(
            x, y, weights=mass, gridsize=gridsize, extents=extents, **kwargs)
        good = np.where(density_mass > 0)
        density[good] = density[good] / density_mass[good]

        if nmin is not None:
            density *= density_mass > nmin

    else:
        # produce a weighted gaussian KDE
        if weights is not None:
            weights = weights[ind[0]]
        elif mass is not None:
            weights = mass[ind[0]]

        density = fast_kde(x, y, weights=weights, gridsize=gridsize,
                           extents=extents, **kwargs)

    try:
        density = SimArray(density, weights.units)
    except AttributeError:
        density = SimArray(density)

    if ret_im:
        return make_contour_plot(density, xs, ys, **kwargs)

    if make_plot:
        make_contour_plot(density, xs, ys, **kwargs)
        if draw_contours:
            make_contour_plot(SimArray(density_mass, mass.units), xs, ys, filled=False,
                              clear=False, colorbar=False, colors='black', scalemin=nmin, nlevels=10)

    return density, xs, ys


def make_contour_plot(arr, xs, ys, x_range=None, y_range=None, nlevels=20,
                      logscale=True, xlogrange=False, ylogrange=False,
                      subplot=False, colorbar=False, ret_im=False, cmap=None,
                      clear=True, legend=False, scalemin=None,levels=None,
                      scalemax=None, filename=None, **kwargs):
    """
    Plot a contour plot of grid *arr* corresponding to bin centers
    specified by *xs* and *ys*.  Labels the axes and colobar with
    proper units taken from x

    Called by :func:`~pynbody.plot.generic.hist2d` and
    :func:`~pynbody.plot.generic.gauss_density`.

    **Input**:

       *arr*: 2D array to plot

       *xs*: x-coordinates of bins

       *ys*: y-coordinates of bins

    **Optional Keywords**:

       *x_range*: list, array, or tuple (default = None)
         size(x_range) must be 2. Specifies the X range.

       *y_range*: tuple (default = None)
         size(y_range) must be 2. Specifies the Y range.

       *xlogrange*: boolean (default = False)
         whether the x-axis should have a log scale

       *ylogrange*: boolean (default = False)
         whether the y-axis should have a log scale

       *nlevels*: int (default = 20)
         number of levels to use for the contours

       *logscale*: boolean (default = True)
         whether to use log or linear spaced contours

       *colorbar*: boolean (default = False)
         draw a colorbar

       *scalemin*: float (default = arr.min())
         minimum value to use for the color scale

       *scalemax*: float (default = arr.max())
         maximum value to use for the color scale
    """

    from matplotlib import colors

    if not subplot:
        import matplotlib.pyplot as plt
    else:
        plt = subplot

    if scalemin is None:
        scalemin = np.min(arr[arr > 0])
    if scalemax is None:
        scalemax = np.max(arr[arr > 0])
    arr[arr < scalemin] = scalemin
    arr[arr > scalemax] = scalemax

    if 'norm' in kwargs:
        del(kwargs['norm'])

    if logscale:
        try:
            levels = np.logspace(np.log10(scalemin),
                                 np.log10(scalemax), nlevels)
            cont_color = colors.LogNorm()
        except ValueError:
            raise ValueError(
                'crazy contour levels -- try specifying the *levels* keyword or use a linear scale')

            return

        if arr.units != NoUnit() and arr.units != 1:
            cb_label = '$log_{10}(' + arr.units.latex() + ')$'
        else:
            cb_label = '$log_{10}(N)$'
    else:
        levels = np.linspace(scalemin,
                             scalemax, nlevels)
        cont_color = None

        if arr.units != NoUnit() and arr.units != 1:
            cb_label = '$' + arr.units.latex() + '$'
        else:
            cb_label = '$N$'

    if not subplot and clear:
        plt.clf()

    if ret_im:
        if logscale:
            arr = np.log10(arr)
        scalemin, scalemax = np.log10((scalemin, scalemax))
        return plt.imshow(arr, origin='down', vmin=scalemin, vmax=scalemax,
                          aspect='auto', cmap=cmap, axes=subplot,
                          # aspect =
                          # np.diff(x_range)/np.diff(y_range),cmap=cmap,
                          extent=[x_range[0], x_range[1], y_range[0], y_range[1]])
    cs = plt.contourf(
        xs, ys, arr, levels, norm=cont_color, cmap=cmap, **kwargs)

    if 'xlabel' in kwargs:
        xlabel = kwargs['xlabel']
    else:
        try:
            if xlogrange:
                xlabel = r'' + '$log_{10}(' + xs.units.latex() + ')$'
            else:
                xlabel = r'' + '$x/' + xs.units.latex() + '$'
        except AttributeError:
            xlabel = None

    if xlabel:
        try:
            if subplot:
                plt.set_xlabel(xlabel)
            else:
                plt.xlabel(xlabel)
        except Exception:
            pass

    if 'ylabel' in kwargs:
        ylabel = kwargs['ylabel']
    else:
        try:
            if ylogrange:
                ylabel = '$log_{10}(' + ys.units.latex() + ')$'
            else:
                ylabel = r'' + '$y/' + ys.units.latex() + '$'
        except AttributeError:
            ylabel = None

    if ylabel:
        try:
            if subplot:
                plt.set_ylabel(ylabel)
            else:
                plt.ylabel(ylabel)
        except Exception:
            pass


#    if not subplot:
#        plt.xlim((x_range[0],x_range[1]))
#        plt.ylim((y_range[0],y_range[1]))

    if colorbar:
        cb = plt.colorbar(cs, format="%.2e").set_label(r'' + cb_label)

    if legend:
        plt.legend(loc=2)

    if (filename):
        if config['verbose']:
            print("Saving " + filename)
        plt.savefig(filename)


def fourier_map(sim, nbins=100, nmin=1000, nphi=100, mmin=1, mmax=7, rmax=10,
                levels=[.01, .05, .1, .2], subplot=None, ret=False, **kwargs):
    """

    Plot an overdensity map generated from a Fourier expansion of the
    particle distribution. A :func:`~pynbody.analysis.profile.Profile`
    is made and passed to :func:`~pynbody.plot.util.inv_fourier` to
    obtain an overdensity map. The map is plotted using the usual
    matplotlib.contour.

    **Input**:

    *sim* :  a :func:`~pynbody.snapshot.SimSnap` object

    **Optional Keywords**:

    *nbins* (100) : number of radial bins to use for the profile

    *nmin* (1000) : minimum number of particles required per bin

    *nphi* (100)  : number of azimuthal bins to use for the map

    *mmin* (1)    : lowest multiplicity Fourier component

    *mmax* (7)    : highest multiplicity Fourier component

    *rmax* (10)   : maximum radius to use when generating the profile

    *levels* [0.01,0.05,0.1,0.2] : tuple of levels for plotting contours

    *subplot* (None) : Axes object on which to plot the contours

    """
    from . import util

    if subplot is None:
        import matplotlib.pylab as plt

    else:
        plt = subplot

    p = pynbody.analysis.profile.Profile(sim, max=rmax, nbins=nbins)
    phi, phi_inv = util.inv_fourier(p, nmin, mmin, mmax, nphi)

    rr, pp = np.meshgrid(p['rbins'], phi)

    xx = (rr * np.cos(pp)).T
    yy = (rr * np.sin(pp)).T

    plt.contour(xx, yy, phi_inv, levels, **kwargs)

    if ret:
        return xx, yy, phi_inv


def prob_plot(x, y, weight, nbins=(100, 100), extent=None, axes=None, **kwargs):
    """

    Make a plot of the probability of y given x, i.e. p(y|x). The
    values are normalized such that the integral along each column is
    one.

    **Input**:

    *x*: primary binning axis

    *y*: secondary binning axis

    *weight*: weights array

    *nbins*: tuple of length 2 specifying the number of bins in each direction

    *extent*: tuple of length 4 speciphysical extent of the axes
     (xmin,xmax,ymin,ymax)

    **Optional Keywords**:

    all optional keywords are passed on to the imshow() command

    """

    import matplotlib.pylab as plt

    assert(len(nbins) == 2)
    grid = np.zeros(nbins)

    if extent is None:
        extent = (min(x), max(x), min(y), max(y))

    xbinedges = np.linspace(extent[0], extent[1], nbins[0] + 1)
    ybinedges = np.linspace(extent[2], extent[3], nbins[1] + 1)

    for i in range(nbins[0]):

        ind = np.where((x > xbinedges[i]) & (x < xbinedges[i + 1]))[0]
        h, bins = np.histogram(
            y[ind], weights=weight[ind], bins=ybinedges, density=True)
        grid[:, i] = h

    if axes is None:
        im = plt.imshow(grid, extent=extent, origin='lower', **kwargs)

    else:
        im = axes.imshow(grid, extent=extent, origin='lower', **kwargs)

    cb = plt.colorbar(im, format='%.2f')
    cb.set_label(r'$P(y|x)$')

    return grid, xbinedges, ybinedges



"""

ionfrac
=======

calculates ionization fractions - NEEDS DOCUMENTATION

"""

import logging
import os

import numpy as np

from pynbody import config

logger = logging.getLogger('pynbody.analysis.ionfrac')

from .interpolate import interpolate3d


def calculate(sim, ion='ovi', mode='old'):
    """

    calculate -- documentation placeholder

    """

    global config
    # ionization fractions calculated for optically thin case with
    # CLOUDY v 10.0.  J. Xavier Prochaska + Joe Hennawi have many
    # helper idl routines for running CLOUDY
    iffile = os.path.join(os.path.dirname(__file__), "ionfracs.npz")
    if os.path.exists(iffile):
        # import data
        logger.info("Loading %s" % iffile)
        ifs = np.load(iffile)
    else:
        raise OSError("ionfracs.npz (Ion Fraction table) not found")

    # allocate temporary metals that we can play with
    # before inlining, the views on the arrays must be standard np.ndarray
    # otherwise the normal numpy macros are not generated
    x_vals = ifs['redshiftvals'].view(np.ndarray)
    y_vals = ifs['tempvals'].view(np.ndarray)
    z_vals = ifs['denvals'].view(np.ndarray)
    vals = ifs[ion + 'if'].view(np.ndarray)
    x = np.zeros(len(sim.gas))
    x[:] = sim.properties['z']
    y = np.log10(sim.gas['temp']).view(np.ndarray)
    z = np.log10(sim.gas['rho'].in_units('m_p cm^-3')).view(np.ndarray)
    n = len(sim.gas)
    n_x_vals = len(x_vals)
    n_y_vals = len(y_vals)
    n_z_vals = len(z_vals)

    # get values off grid to minmax
    x[np.where(x < np.min(x_vals))] = np.min(x_vals)
    x[np.where(x > np.max(x_vals))] = np.max(x_vals)
    y[np.where(y < np.min(y_vals))] = np.min(y_vals)
    y[np.where(y > np.max(y_vals))] = np.max(y_vals)
    z[np.where(z < np.min(z_vals))] = np.min(z_vals)
    z[np.where(z > np.max(z_vals))] = np.max(z_vals)

    # interpolate
    logger.info("Interpolation %s values" % ion)
    result_array = interpolate3d(x, y, z, x_vals, y_vals, z_vals, vals)

    return 10 ** result_array



"""

util
====

Utility functions for the plotting module

"""

import numpy as np
import scipy as sp
import scipy.signal
import scipy.sparse


def fast_kde(x, y, kern_nx=None, kern_ny=None, gridsize=(100, 100),
             extents=None, nocorrelation=False, weights=None, norm = False, **kwargs):
    """
    A faster gaussian kernel density estimate (KDE).  Intended for
    computing the KDE on a regular grid (different use case than
    scipy's original scipy.stats.kde.gaussian_kde()).

    Author: Joe Kington
    License:  MIT License <http://www.opensource.org/licenses/mit-license.php>

    Performs a gaussian kernel density estimate over a regular grid using a
    convolution of the gaussian kernel with a 2D histogram of the data.

    This function is typically several orders of magnitude faster than
    scipy.stats.kde.gaussian_kde for large (>1e7) numbers of points and
    produces an essentially identical result.

    **Input**:

        *x*: array
            The x-coords of the input data points

        *y*: array
            The y-coords of the input data points

        *kern_nx*: float
            size (in units of *x*) of the kernel

        *kern_ny*: float
            size (in units of *y*) of the kernel

        *gridsize*: (Nx , Ny) tuple (default: 200x200)
            Size of the output grid

        *extents*: (default: extent of input data) A (xmin, xmax, ymin, ymax)
            tuple of the extents of output grid

        *nocorrelation*: (default: False) If True, the correlation between the
            x and y coords will be ignored when preforming the KDE.

        *weights*: (default: None) An array of the same shape as x & y that
            weighs each sample (x_i, y_i) by each value in weights (w_i).
            Defaults to an array of ones the same size as x & y.

        *norm*: boolean (default: False)
            If False, the output is only corrected for the kernel. If True,
            the result is normalized such that the integral over the area
            yields 1.

    **Output**:
        A gridded 2D kernel density estimate of the input points.
    """

    #---- Setup --------------------------------------------------------------
    x, y = np.asarray(x), np.asarray(y)
    x, y = np.squeeze(x), np.squeeze(y)

    if x.size != y.size:
        raise ValueError('Input x & y arrays must be the same size!')

    nx, ny = gridsize
    n = x.size

    if weights is None:
        # Default: Weight all points equally
        weights = np.ones(n)
    else:
        weights = np.squeeze(np.asarray(weights))
        if weights.size != x.size:
            raise ValueError('Input weights must be an array of the same size'
                             ' as input x & y arrays!')

    # Default extents are the extent of the data
    if extents is None:
        xmin, xmax = x.min(), x.max()
        ymin, ymax = y.min(), y.max()
    else:
        xmin, xmax, ymin, ymax = list(map(float, extents))

    dx = (xmax - xmin) / (nx - 1)
    dy = (ymax - ymin) / (ny - 1)

    #---- Preliminary Calculations -------------------------------------------

    # First convert x & y over to pixel coordinates
    # (Avoiding np.digitize due to excessive memory usage!)
    xyi = np.vstack((x, y)).T
    xyi -= [xmin, ymin]
    xyi /= [dx, dy]
    xyi = np.floor(xyi, xyi).T

    # Next, make a 2D histogram of x & y
    # Avoiding np.histogram2d due to excessive memory usage with many points
    grid = sp.sparse.coo_matrix((weights, xyi), shape=(nx, ny)).toarray()

    # Calculate the covariance matrix (in pixel coords)
    cov = np.cov(xyi)

    if nocorrelation:
        cov[1, 0] = 0
        cov[0, 1] = 0

    # Scaling factor for bandwidth
    scotts_factor = np.power(n, -1.0 / 6)  # For 2D

    #---- Make the gaussian kernel -------------------------------------------

    # First, determine how big the kernel needs to be
    std_devs = np.diag(np.sqrt(cov))

    if kern_nx is None or kern_ny is None:
        kern_nx, kern_ny = np.round(scotts_factor * 2 * np.pi * std_devs)

    else:
        kern_nx = np.round(kern_nx / dx)
        kern_ny = np.round(kern_ny / dy)

    # Determine the bandwidth to use for the gaussian kernel
    inv_cov = np.linalg.inv(cov * scotts_factor ** 2)

    # x & y (pixel) coords of the kernel grid, with <x,y> = <0,0> in center
    xx = np.arange(kern_nx, dtype=np.float_) - kern_nx / 2.0
    yy = np.arange(kern_ny, dtype=np.float_) - kern_ny / 2.0
    xx, yy = np.meshgrid(xx, yy)

    # Then evaluate the gaussian function on the kernel grid
    kernel = np.vstack((xx.flatten(), yy.flatten()))
    kernel = np.dot(inv_cov, kernel) * kernel
    kernel = np.sum(kernel, axis=0) / 2.0
    kernel = np.exp(-kernel)
    kernel = kernel.reshape((int(kern_ny), int(kern_nx)))

    #---- Produce the kernel density estimate --------------------------------

    # Convolve the gaussian kernel with the 2D histogram, producing a gaussian
    # kernel density estimate on a regular grid
    grid = sp.signal.convolve2d(grid, kernel, mode='same', boundary='fill').T

    # Normalization factor to divide result by so that units are in the same
    # units as scipy.stats.kde.gaussian_kde's output.
    norm_factor = 2 * np.pi * cov * scotts_factor ** 2
    norm_factor = np.linalg.det(norm_factor)
    #norm_factor = n * dx * dy * np.sqrt(norm_factor)
    norm_factor = np.sqrt(norm_factor)

    if norm:
        norm_factor *= n * dx * dy

    # Normalize the result
    grid /= norm_factor

    return grid


def inv_fourier(p, nmin=1000, mmin=1, mmax=7, nphi=100):
    """

    Invert a profile with fourier coefficients to yield an overdensity
    map.

    **Inputs:**

    *p* : a :func:`~pynbody.analysis.profile.Profile` object

    **Optional Keywords:**

    *nmin* (1000) : minimum number of particles required per bin

    *mmin* (1)    : lowest multiplicity Fourier component

    *mmax* (7)    : highest multiplicity Fourier component

    *nphi* (100)  : number of azimuthal bins to use for the map

    """

    phi_hist = np.zeros((len(p['rbins']), nphi))
    phi = np.linspace(-np.pi, np.pi, nphi)
    rbins = p['rbins']

    for i in range(len(rbins)):
        if p['n'][i] > nmin:
            for m in range(mmin, mmax):
                phi_hist[i, :] = phi_hist[i,:] + p['fourier']['c'][m, i]*np.exp(1j*m*phi)

    return phi, phi_hist



"""Gravity Tree. Builds tree based on pkdgrav2"""

try:
    from .. import pkdgrav
except ImportError:
    # pkdgrav2 never works at the moment, and the warning below
    # confuses/annoys people
    pass
    #import warnings
    #warnings.warn("Unable to import PKDGrav gravity solver. Most likely this means either that your installation is broken, or that you are running python inside the pynbody distribution directory, in which case python cannot see the installed version. However, it also doesn't matter unless you want to use the tree gravity solver.",RuntimeWarning)

from time import process_time

import numpy as np

from .. import config


class GravTree:

    def __init__(self, pos, mass, rs, eps=None, leafsize=16):

        start = process_time()
        self.tree = pkdgrav.pkdPythonInitialize(
            pos, mass, eps, rs, int(leafsize))
        end = process_time()
        if config['verbose']:
            print('Tree build done in %5.3g s' % (end - start))

        self.derived = True
        self.flags = {'WRITEABLE': False}

    def calc(self, vec_pos, eps=None):
        accel = np.zeros((len(vec_pos), 3))
        pot = np.zeros(len(vec_pos))
        if config['verbose']:
            print('Calculating Gravity')

        start = process_time()
        pkdgrav.pkdPythonDoGravity(self.tree, accel, pot, theta=0.55)
        end = process_time()
        if config['verbose']:
            print('Gravity calculated in %5.3g s' % (end - start))

        return accel, pot

    def __del__(self):
        if hasattr(self, 'tree'):
            pkdgrav.free(self.tree)



r"""
Defines a docutils directive for inserting inheritance diagrams.

Provide the directive with one or more classes or modules (separated
by whitespace).  For modules, all of the classes in that module will
be used.

Example::

   Given the following classes:

   class A: pass
   class B(A): pass
   class C(A): pass
   class D(B, C): pass
   class E(B): pass

   .. inheritance-diagram: D E

   Produces a graph like the following:

               A
              / \
             B   C
            / \ /
           E   D

The graph is inserted as a PNG+image map into HTML and a PDF in
LaTeX.
"""

import inspect
import os
import re
import subprocess

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

from docutils.nodes import Body, Element
from docutils.parsers.rst import directives
from sphinx.roles import xfileref_role


def my_import(name):
    """Module importer - taken from the python documentation.

    This function allows importing names with dots in them."""

    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

class DotException(Exception):
    pass

class InheritanceGraph(object):
    """
    Given a list of classes, determines the set of classes that
    they inherit from all the way to the root "object", and then
    is able to generate a graphviz dot graph from them.
    """
    def __init__(self, class_names, show_builtins=False):
        """
        *class_names* is a list of child classes to show bases from.

        If *show_builtins* is True, then Python builtins will be shown
        in the graph.
        """
        self.class_names = class_names
        self.classes = self._import_classes(class_names)
        self.all_classes = self._all_classes(self.classes)
        if len(self.all_classes) == 0:
            raise ValueError("No classes found for inheritance diagram")
        self.show_builtins = show_builtins

    py_sig_re = re.compile(r'''^([\w.]*\.)?    # class names
                           (\w+)  \s* $        # optionally arguments
                           ''', re.VERBOSE)

    def _import_class_or_module(self, name):
        """
        Import a class using its fully-qualified *name*.
        """
        try:
            path, base = self.py_sig_re.match(name).groups()
        except Exception as e:
            raise ValueError(
                "Invalid class or module '%s' specified for inheritance diagram" % name
            ) from e
        fullname = (path or '') + base
        path = (path and path.rstrip('.'))
        if not path:
            path = base
        try:
            module = __import__(path, None, None, [])
            # We must do an import of the fully qualified name.  Otherwise if a
            # subpackage 'a.b' is requested where 'import a' does NOT provide
            # 'a.b' automatically, then 'a.b' will not be found below.  This
            # second call will force the equivalent of 'import a.b' to happen
            # after the top-level import above.
            my_import(fullname)

        except ImportError:
            raise ValueError(
                "Could not import class or module '%s' specified for inheritance diagram" % name)

        try:
            todoc = module
            for comp in fullname.split('.')[1:]:
                todoc = getattr(todoc, comp)
        except AttributeError:
            raise ValueError(
                "Could not find class or module '%s' specified for inheritance diagram" % name)

        # If a class, just return it
        if inspect.isclass(todoc):
            return [todoc]
        elif inspect.ismodule(todoc):
            classes = []
            for cls in todoc.__dict__.values():
                if inspect.isclass(cls) and cls.__module__ == todoc.__name__:
                    classes.append(cls)
            return classes
        raise ValueError(
            "'%s' does not resolve to a class or module" % name)

    def _import_classes(self, class_names):
        """
        Import a list of classes.
        """
        classes = []
        for name in class_names:
            classes.extend(self._import_class_or_module(name))
        return classes

    def _all_classes(self, classes):
        """
        Return a list of all classes that are ancestors of *classes*.
        """
        all_classes = {}

        def recurse(cls):
            all_classes[cls] = None
            for c in cls.__bases__:
                if c not in all_classes:
                    recurse(c)

        for cls in classes:
            recurse(cls)

        return all_classes.keys()

    def class_name(self, cls, parts=0):
        """
        Given a class object, return a fully-qualified name.  This
        works for things I've tested in matplotlib so far, but may not
        be completely general.
        """
        module = cls.__module__
        if module == '__builtin__':
            fullname = cls.__name__
        else:
            fullname = "%s.%s" % (module, cls.__name__)
        if parts == 0:
            return fullname
        name_parts = fullname.split('.')
        return '.'.join(name_parts[-parts:])

    def get_all_class_names(self):
        """
        Get all of the class names involved in the graph.
        """
        return [self.class_name(x) for x in self.all_classes]

    # These are the default options for graphviz
    default_graph_options = {
        "rankdir": "LR",
        "size": '"8.0, 12.0"'
        }
    default_node_options = {
        "shape": "box",
        "fontsize": 10,
        "height": 0.25,
        "fontname": "Vera Sans, DejaVu Sans, Liberation Sans, Arial, Helvetica, sans",
        "style": '"setlinewidth(0.5)"'
        }
    default_edge_options = {
        "arrowsize": 0.5,
        "style": '"setlinewidth(0.5)"'
        }

    def _format_node_options(self, options):
        return ','.join(["%s=%s" % x for x in options.items()])
    def _format_graph_options(self, options):
        return ''.join(["%s=%s;\n" % x for x in options.items()])

    def generate_dot(self, fd, name, parts=0, urls={},
                     graph_options={}, node_options={},
                     edge_options={}):
        """
        Generate a graphviz dot graph from the classes that
        were passed in to __init__.

        *fd* is a Python file-like object to write to.

        *name* is the name of the graph

        *urls* is a dictionary mapping class names to http urls

        *graph_options*, *node_options*, *edge_options* are
        dictionaries containing key/value pairs to pass on as graphviz
        properties.
        """
        g_options = self.default_graph_options.copy()
        g_options.update(graph_options)
        n_options = self.default_node_options.copy()
        n_options.update(node_options)
        e_options = self.default_edge_options.copy()
        e_options.update(edge_options)

        fd.write('digraph %s {\n' % name)
        fd.write(self._format_graph_options(g_options))

        for cls in self.all_classes:
            if not self.show_builtins and cls in __builtins__.values():
                continue

            name = self.class_name(cls, parts)

            # Write the node
            this_node_options = n_options.copy()
            url = urls.get(self.class_name(cls))
            if url is not None:
                this_node_options['URL'] = '"%s"' % url
            fd.write('  "%s" [%s];\n' %
                     (name, self._format_node_options(this_node_options)))

            # Write the edges
            for base in cls.__bases__:
                if not self.show_builtins and base in __builtins__.values():
                    continue

                base_name = self.class_name(base, parts)
                fd.write('  "%s" -> "%s" [%s];\n' %
                         (base_name, name,
                          self._format_node_options(e_options)))
        fd.write('}\n')

    def run_dot(self, args, name, parts=0, urls={},
                graph_options={}, node_options={}, edge_options={}):
        """
        Run graphviz 'dot' over this graph, returning whatever 'dot'
        writes to stdout.

        *args* will be passed along as commandline arguments.

        *name* is the name of the graph

        *urls* is a dictionary mapping class names to http urls

        Raises DotException for any of the many os and
        installation-related errors that may occur.
        """
        try:
            dot = subprocess.Popen(['dot'] + list(args),
                                   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                                   close_fds=True)
        except OSError as e:
            raise DotException("Could not execute 'dot'. Are you sure you have 'graphviz' installed?") from e
        except ValueError as e:
            raise DotException("'dot' called with invalid arguments") from e
        except Exception as e:
            raise DotException("Unexpected error calling 'dot'") from e

        self.generate_dot(dot.stdin, name, parts, urls, graph_options,
                          node_options, edge_options)
        dot.stdin.close()
        result = dot.stdout.read()
        returncode = dot.wait()
        if returncode != 0:
            raise DotException("'dot' returned the errorcode %d" % returncode)
        return result

class inheritance_diagram(Body, Element):
    """
    A docutils node to use as a placeholder for the inheritance
    diagram.
    """
    pass

def inheritance_diagram_directive(name, arguments, options, content, lineno,
                                  content_offset, block_text, state,
                                  state_machine):
    """
    Run when the inheritance_diagram directive is first encountered.
    """
    node = inheritance_diagram()

    class_names = arguments

    # Create a graph starting with the list of classes
    graph = InheritanceGraph(class_names)

    # Create xref nodes for each target of the graph's image map and
    # add them to the doc tree so that Sphinx can resolve the
    # references to real URLs later.  These nodes will eventually be
    # removed from the doctree after we're done with them.
    for name in graph.get_all_class_names():
        refnodes, x = xfileref_role(
            'class', ':class:`%s`' % name, name, 0, state)
        node.extend(refnodes)
    # Store the graph object so we can use it to generate the
    # dot file later
    node['graph'] = graph
    # Store the original content for use as a hash
    node['parts'] = options.get('parts', 0)
    node['content'] = " ".join(class_names)
    return [node]

def get_graph_hash(node):
    return md5(node['content'] + str(node['parts'])).hexdigest()[-10:]

def html_output_graph(self, node):
    """
    Output the graph for HTML.  This will insert a PNG with clickable
    image map.
    """
    graph = node['graph']
    parts = node['parts']

    graph_hash = get_graph_hash(node)
    name = "inheritance%s" % graph_hash
    path = '_images'
    dest_path = os.path.join(setup.app.builder.outdir, path)
    if not os.path.exists(dest_path):
        os.makedirs(dest_path)
    png_path = os.path.join(dest_path, name + ".png")
    path = setup.app.builder.imgpath

    # Create a mapping from fully-qualified class names to URLs.
    urls = {}
    for child in node:
        if child.get('refuri') is not None:
            urls[child['reftitle']] = child.get('refuri')
        elif child.get('refid') is not None:
            urls[child['reftitle']] = '#' + child.get('refid')

    # These arguments to dot will save a PNG file to disk and write
    # an HTML image map to stdout.
    image_map = graph.run_dot(['-Tpng', '-o%s' % png_path, '-Tcmapx'],
                              name, parts, urls)
    return ('<img src="%s/%s.png" usemap="#%s" class="inheritance"/>%s' %
            (path, name, name, image_map))

def latex_output_graph(self, node):
    """
    Output the graph for LaTeX.  This will insert a PDF.
    """
    graph = node['graph']
    parts = node['parts']

    graph_hash = get_graph_hash(node)
    name = "inheritance%s" % graph_hash
    dest_path = os.path.abspath(os.path.join(setup.app.builder.outdir, '_images'))
    if not os.path.exists(dest_path):
        os.makedirs(dest_path)
    pdf_path = os.path.abspath(os.path.join(dest_path, name + ".pdf"))

    graph.run_dot(['-Tpdf', '-o%s' % pdf_path],
                  name, parts, graph_options={'size': '"6.0,6.0"'})
    return '\n\\includegraphics{%s}\n\n' % pdf_path

def visit_inheritance_diagram(inner_func):
    """
    This is just a wrapper around html/latex_output_graph to make it
    easier to handle errors and insert warnings.
    """
    def visitor(self, node):
        try:
            content = inner_func(self, node)
        except DotException, e:
            # Insert the exception as a warning in the document
            warning = self.document.reporter.warning(str(e), line=node.line)
            warning.parent = node
            node.children = [warning]
        else:
            source = self.document.attributes['source']
            self.body.append(content)
            node.children = []
    return visitor

def do_nothing(self, node):
    pass

def setup(app):
    setup.app = app
    setup.confdir = app.confdir

    app.add_node(
        inheritance_diagram,
        latex=(visit_inheritance_diagram(latex_output_graph), do_nothing),
        html=(visit_inheritance_diagram(html_output_graph), do_nothing))
    app.add_directive(
        'inheritance-diagram', inheritance_diagram_directive,
        False, (1, 100, 0), parts = directives.nonnegative_int)



#cython: embedsignature=True

cimport cython

import numpy as np

from pynbody import array, config, openmp, units
from pynbody.util import get_eps

cimport numpy as np

DTYPE = np.double

ctypedef fused DTYPE_t:
    np.float32_t
    np.float64_t

cdef extern from "math.h" nogil:
      double sqrt(double)
      float sqrt(float)


@cython.cdivision(True)
@cython.boundscheck(False)
def direct(f, np.ndarray[DTYPE_t, ndim=2] ipos, eps=None, int num_threads = 0):

    from cython.parallel cimport prange

    global config


    if num_threads == 0 :
        num_threads = int(config["number_of_threads"])

    if num_threads < 0:
        num_threads = openmp.get_cpus()

    if num_threads > openmp.get_cpus() :
        num_threads = openmp.get_cpus()

    openmp.set_threads(num_threads)

    if eps is None:
        eps = get_eps(f)

    cdef unsigned int nips = len(ipos)
    cdef np.ndarray[DTYPE_t, ndim=2] m_by_r2 = np.zeros((nips,3), dtype = ipos.dtype)
    cdef np.ndarray[DTYPE_t, ndim=1] m_by_r = np.zeros(nips, dtype = ipos.dtype)
    cdef np.ndarray[DTYPE_t, ndim=2] pos = f['pos'].view(np.ndarray)
    cdef np.ndarray[DTYPE_t, ndim=1] mass = f['mass'].view(np.ndarray)
    cdef unsigned int n = len(mass)
    cdef np.ndarray[DTYPE_t, ndim=1] epssq = eps * eps

    cdef unsigned int pi, i
    cdef double dx, dy, dz, mass_i, epssq_i, drsoft, drsoft3

    for pi in prange(nips, nogil=True, schedule='static'):
        for i in range(n):
            mass_i = mass[i]
            epssq_i = epssq[i]
            dx = ipos[pi,0] - pos[i,0]
            dy = ipos[pi,1] - pos[i,1]
            dz = ipos[pi,2] - pos[i,2]
            drsoft = 1.0/sqrt(dx*dx + dy*dy + dz*dz + epssq_i)
            drsoft3 = drsoft*drsoft*drsoft
            m_by_r[pi] += mass_i * drsoft
            m_by_r2[pi,0] += mass_i*dx * drsoft3
            m_by_r2[pi,1] += mass_i*dy * drsoft3
            m_by_r2[pi,2] += mass_i*dz * drsoft3


    m_by_r = m_by_r.view(array.SimArray)
    m_by_r2 = m_by_r2.view(array.SimArray)
    m_by_r.units = f['mass'].units/f['pos'].units
    m_by_r2.units = f['mass'].units/f['pos'].units**2

    m_by_r*=units.G
    m_by_r2*=units.G

    return -m_by_r, -m_by_r2



"""

gas
===

Functions for plotting gas quantities

"""

import logging

import matplotlib.pyplot as plt
import numpy as np

from ..analysis import angmom, profile
from ..units import Unit
from .generic import hist2d

logger = logging.getLogger('pynbody.plot.gas')


def rho_T(sim, rho_units=None, rho_range=None, t_range=None, two_phase='split', **kwargs):
    """
    Plot Temperature vs. Density for the gas particles in the snapshot.

    **Optional keywords:**

       *rho_units*: specify the density units (default is the same units as the current 'rho' array)

       *t_range*: list, array, or tuple
          ``size(t_range)`` must be 2. Specifies the temperature range.

       *rho_range:* tuple
          ``size(rho_range)`` must be 2. Specifies the density range.

       *two_phase*: if two-phase particles are detected, either plot each phase separately ('split'), or merge them ('merge')

    See :func:`~pynbody.plot.generic.hist2d` for other plotting keyword options


    """
    if rho_units is None:
        rho_units = sim.gas['rho'].units

    if t_range is not None:
        kwargs['y_range'] = t_range
        assert len(kwargs['y_range']) == 2

    if rho_range is not None:
        kwargs['x_range'] = rho_range
        assert len(kwargs['x_range']) == 2
    else:
        rho_range = False

    if 'xlabel' in kwargs:
        xlabel = kwargs['xlabel']
        del kwargs['xlabel']
    else:
        xlabel = r'log$_{10}$($\rho$/$' + Unit(rho_units).latex() + '$)'

    if 'ylabel' in kwargs:
        ylabel = kwargs['ylabel']
        del kwargs['ylabel']
    else:
        ylabel = r'log$_{10}$(T/$' + sim.gas['temp'].units.latex() + '$)'

    if 'Tinc' in sim.loadable_keys() and two_phase == 'merge':
        return hist2d(sim.gas['rho'].in_units(rho_units),sim.gas['Tinc'],
                      xlogrange=True,ylogrange=True,xlabel=xlabel,
                      ylabel=ylabel, **kwargs)

    if 'uHot' in sim.loadable_keys() and 'MassHot' in sim.loadable_keys() and two_phase == 'split':
        E = sim.g['uHot']*sim.g['MassHot']+sim.g['u']*(sim.g['mass']-sim.g['MassHot'])
        rho = np.concatenate((sim.g['rho'].in_units(rho_units)*E/(sim.g['mass']*sim.g['u']),
            sim.g['rho'].in_units(rho_units)*E/(sim.g['mass']*sim.g['uHot'])))
        temp = np.concatenate((sim.g['temp'], sim.g['temp']/sim.g['u']*sim.g['uHot']))
        temp = temp[np.where(np.isfinite(rho))]
        rho = rho[np.where(np.isfinite(rho))]
        return hist2d(rho, temp, xlogrange=True,ylogrange=True,xlabel=xlabel,
                ylabel=ylabel, **kwargs)

    return hist2d(sim.gas['rho'].in_units(rho_units),sim.gas['temp'],
                  xlogrange=True,ylogrange=True,xlabel=xlabel,
                  ylabel=ylabel, **kwargs)



def temp_profile(sim, center=True, r_units='kpc', bin_spacing='equaln',
                 clear=True, filename=None, **kwargs):
    """

    Centre on potential minimum, align so that the disk is in the
    x-y plane, then plot the temperature profile as a
    function of radius.

    """

    if center:
        angmom.sideon(sim)

    if 'min' in kwargs:
        min_r = kwargs['min']
    else:
        min_r = sim['r'].min()
    if 'max' in kwargs:
        max_r = kwargs['max']
    else:
        max_r = sim['r'].max()

    pro = profile.Profile(sim.gas, type=bin_spacing, rmin =min_r, rmax =max_r)

    r = pro['rbins'].in_units(r_units)
    tempprof = pro['temp']

    if clear:
        plt.clf()

    plt.semilogy(r, tempprof)

    plt.xlabel("r / $" + r.units.latex() + "$")
    plt.ylabel("Temperature [K]")
    if (filename):
        logger.info("Saving %s", filename)
        plt.savefig(filename)



"""

filt
====

Defines and implements 'filters' which allow abstract subsets
of data to be specified.

See the `filter tutorial
<http://pynbody.github.io/pynbody/tutorials/filters.html>`_ for some
sample usage.

"""


import pickle

import numpy as np

from . import _util, family, units


class Filter:

    def __init__(self):
        self._descriptor = "filter"
        pass

    def where(self, sim):
        return np.where(self(sim))

    def __call__(self, sim):
        return np.ones(len(sim), dtype=bool)

    def __and__(self, f2):
        return And(self, f2)

    def __invert__(self):
        return Not(self)

    def __or__(self, f2):
        return Or(self, f2)

    def __repr__(self):
        return "Filter()"

    def __hash__(self):
        return hash(pickle.dumps(self))

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            else:
                equal = other.__dict__[k]==v
                if isinstance(equal, np.ndarray):
                    equal = equal.all()
                if not equal:
                    return False

        return True


class FamilyFilter(Filter):
    def __init__(self, family_):
        assert isinstance(family_, family.Family)
        self._descriptor = family_.name
        self.family = family_

    def __repr__(self):
        return "FamilyFilter("+self.family.name+")"

    def __call__(self, sim):
        slice_ = sim._get_family_slice(self.family)
        flags = np.zeros(len(sim), dtype=bool)
        flags[slice_] = True
        return flags

class And(Filter):

    def __init__(self, f1, f2):
        self._descriptor = f1._descriptor + "&" + f2._descriptor
        self.f1 = f1
        self.f2 = f2

    def __call__(self, sim):
        return self.f1(sim) * self.f2(sim)

    def __repr__(self):
        return "(" + repr(self.f1) + " & " + repr(self.f2) + ")"


class Or(Filter):

    def __init__(self, f1, f2):
        self._descriptor = f1._descriptor + "|" + f2._descriptor
        self.f1 = f1
        self.f2 = f2

    def __call__(self, sim):
        return self.f1(sim) + self.f2(sim)

    def __repr__(self):
        return "(" + repr(self.f1) + " | " + repr(self.f2) + ")"


class Not(Filter):

    def __init__(self, f):
        self._descriptor = "~" + f._descriptor
        self.f = f

    def __call__(self, sim):
        x = self.f(sim)
        return np.logical_not(x)

    def __repr__(self):
        return "~" + repr(self.f)


class Sphere(Filter):

    """
    Return particles that are within `radius` of the point `cen`.

    Inputs:
    -------

    *radius* : extent of the sphere. Can be a number or a string specifying the units.

    *cen* : center of the sphere. default = (0,0,0)
    """

    def __init__(self, radius, cen=(0, 0, 0)):
        self._descriptor = "sphere"
        self.cen = np.asarray(cen)
        if self.cen.shape != (3,):
            raise ValueError("Centre must be length 3 array")

        if isinstance(radius, str):
            radius = units.Unit(radius)

        self.radius = radius

    def __call__(self, sim):
        wrap = -1.0

        with sim.immediate_mode:
            pos = sim['pos']

        cen, radius = self._get_cen_and_radius_as_float(pos)

        if 'boxsize' in sim.properties:
            wrap = sim.properties['boxsize']

        if units.is_unit_like(wrap):
            wrap = float(wrap.in_units(pos.units,**pos.conversion_context()))

        return _util._sphere_selection(np.asarray(pos),np.asarray(cen,dtype=pos.dtype),radius,wrap)

    def _get_cen_and_radius_as_float(self, pos):
        radius = self.radius
        cen = self.cen
        if units.has_units(cen):
            cen = cen.in_units(pos.units)
        if units.is_unit_like(radius):
            radius = float(radius.in_units(pos.units,
                                           **pos.conversion_context()))
        return cen, radius

    def where(self, sim):
        if hasattr(sim, "kdtree"):
            cen, radius = self._get_cen_and_radius_as_float(sim["pos"])
            return (np.sort(sim.kdtree.particles_in_sphere(cen, radius)),)
        else:
            return super().where(sim)

    def __repr__(self):
        if units.is_unit(self.radius):

            return f"Sphere('{str(self.radius)}', {repr(self.cen)})"
        else:
            return f"Sphere({self.radius:.2e}, {repr(self.cen)})"


class Cuboid(Filter):

    """Create a cube with specified edge coordinates. If any of the cube
    coordinates `x1`, `y1`, `z1`, `x2`, `y2`, `z2` are not specified
    they are determined as `y1=x1;` `z1=x1;` `x2=-x1;` `y2=-y1;`
    `z2=-z1`.

    """

    def __init__(self, x1, y1=None, z1=None, x2=None, y2=None, z2=None):

        self._descriptor = "cube"
        x1, y1, z1, x2, y2, z2 = (
            units.Unit(x) if isinstance(x, str) else x for x in (x1, y1, z1, x2, y2, z2))
        if y1 is None:
            y1 = x1
        if z1 is None:
            z1 = x1
        if x2 is None:
            x2 = -x1
        if y2 is None:
            y2 = -y1
        if z2 is None:
            z2 = -z1
        self.x1, self.y1, self.z1, self.x2, self.y2, self.z2 = x1, y1, z1, x2, y2, z2

    def __call__(self, sim):
        return self._get_mask(sim['x'], sim['y'], sim['z'], self._get_boundaries(sim))

    def _get_mask(self, x, y, z, boundaries):
        x1, y1, z1, x2, y2, z2 = boundaries
        return ((x > x1) * (x < x2) * (y > y1) * (y < y2) * (z > z1) * (z < z2))

    def _get_boundaries(self, sim):
        return [x.in_units(sim["pos"].units, **sim["pos"].conversion_context())
                if units.is_unit_like(x) else x
                for x in (self.x1, self.y1, self.z1, self.x2, self.y2, self.z2)]

    def _get_bounding_sphere(self, cuboid_boundaries):
        x1, y1, z1, x2, y2, z2 = cuboid_boundaries
        return ((x1 + x2) / 2, (y1 + y2) / 2, (z1 + z2) / 2), np.sqrt(
            (x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2) / 2

    def where(self, sim):
        if hasattr(sim, "kdtree"):
            # KDTree doesn't currently natively find cuboid regions, so we get the bounding sphere
            # and the search for the cuboid within that
            cuboid_boundaries = self._get_boundaries(sim)
            cen, radius = self._get_bounding_sphere(cuboid_boundaries)
            in_sphere_particles = sim.kdtree.particles_in_sphere(cen, radius)
            x, y, z = sim["pos"][in_sphere_particles].T
            mask = self._get_mask(x, y, z, cuboid_boundaries)
            return np.sort(in_sphere_particles[mask]),
        else:
            return super().where(sim)

    def __repr__(self):
        x1, y1, z1, x2, y2, z2 = ("'%s'" % str(x)
                                  if units.is_unit_like(x) else x
                                  for x in (self.x1, self.y1, self.z1, self.x2, self.y2, self.z2))
        return f"Cuboid({x1}, {y1}, {z1}, {x2}, {y2}, {z2})"


class Disc(Filter):

    """
    Return particles that are within a disc of extent `radius` and
    thickness `height` centered on `cen`.
    """

    def __init__(self, radius, height, cen=(0, 0, 0)):
        self._descriptor = "disc"
        self.cen = np.asarray(cen)
        if self.cen.shape != (3,):
            raise ValueError("Centre must be length 3 array")

        if isinstance(radius, str):
            radius = units.Unit(radius)

        if isinstance(height, str):
            height = units.Unit(height)

        self.radius = radius
        self.height = height

    def __call__(self, sim):
        radius = self.radius
        height = self.height

        if units.is_unit_like(radius):
            radius = float(
                radius.in_units(sim["pos"].units, **sim["pos"].conversion_context()))
        if units.is_unit_like(height):
            height = float(
                height.in_units(sim["pos"].units, **sim["pos"].conversion_context()))
        distance = (((sim["pos"] - self.cen)[:, :2]) ** 2).sum(axis=1)
        return (distance < radius ** 2) * (np.abs(sim["z"] - self.cen[2]) < height)

    def __repr__(self):
        radius = self.radius
        height = self.height

        radius, height = (
            ("'%s'" % str(x) if units.is_unit_like(x) else '%.2e' % x) for x in (radius, height))

        return f"Disc({radius}, {height}, {repr(self.cen)})"


class BandPass(Filter):

    """
    Return particles whose property `prop` is within `min` and `max`,
    which can be specified as unit strings.
    """

    def __init__(self, prop, min, max):
        self._descriptor = "bandpass_" + prop

        if isinstance(min, str):
            min = units.Unit(min)

        if isinstance(max, str):
            max = units.Unit(max)

        self._prop = prop
        self._min = min
        self._max = max

    def __call__(self, sim):
        min_ = self._min
        max_ = self._max
        prop = self._prop

        if units.is_unit_like(min_):
            min_ = float(
                min_.in_units(sim[prop].units, **sim.conversion_context()))
        if units.is_unit_like(max_):
            max_ = float(
                max_.in_units(sim[prop].units, **sim.conversion_context()))

        return ((sim[prop] > min_) * (sim[prop] < max_))

    def __repr__(self):
        min_, max_ = (("'%s'" % str(x) if units.is_unit_like(
            x) else '%.2e' % x) for x in (self._min, self._max))
        return f"BandPass('{self._prop}', {min_}, {max_})"


class HighPass(Filter):

    """
    Return particles whose property `prop` exceeds `min`, which can be
    specified as a unit string.
    """

    def __init__(self, prop, min):
        self._descriptor = "highpass_" + prop

        if isinstance(min, str):
            min = units.Unit(min)

        self._prop = prop
        self._min = min

    def __call__(self, sim):
        min_ = self._min

        prop = self._prop

        if units.is_unit_like(min_):
            min_ = float(
                min_.in_units(sim[prop].units, **sim.conversion_context()))

        return (sim[prop] > min_)

    def __repr__(self):
        min = ("'%s'" % str(self._min) if units.is_unit_like(
            self._min) else '%.2e' % self._min)
        return f"HighPass('{self._prop}', {min})"


class LowPass(Filter):

    """Return particles whose property `prop` is less than `max`, which can be
    specified as a unit string.
    """

    def __init__(self, prop, max):
        self._descriptor = "lowpass_" + prop

        if isinstance(max, str):
            max = units.Unit(max)

        self._prop = prop
        self._max = max

    def __call__(self, sim):
        max_ = self._max

        prop = self._prop

        if units.is_unit_like(max_):
            max_ = float(
                max_.in_units(sim[prop].units, **sim.conversion_context()))

        return (sim[prop] < max_)

    def __repr__(self):
        max = ("'%s'" % str(self._max) if isinstance(
            self._max, units.UnitBase) else '%.2e' % self._max)
        return f"LowPass('{self._prop}', {max})"


def Annulus(r1, r2, cen=(0, 0, 0)):
    """
    Convenience function that returns a filter which selects particles
    in between two spheres specified by radii `r1` and `r2` centered
    on `cen`.
    """

    x = Sphere(max(r1, r2), cen) & ~Sphere(min(r1, r2), cen)
    x._descriptor = "annulus"
    return x


def SolarNeighborhood(r1=units.Unit("5 kpc"), r2=units.Unit("10 kpc"), height=units.Unit("2 kpc"), cen=(0, 0, 0)):
    """
    Convenience function that returns a filter which selects particles
    in a disc between radii `r1` and `r2` and thickness `height`.
    """

    x = Disc(max(r1, r2), height, cen) & ~Disc(min(r1, r2), height, cen)
    x._descriptor = "Solar Neighborhood"
    return x



from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

import pynbody

test_folder = Path(__file__).parent

def setup_module():
    global f
    f = pynbody.load("testdata/g15784.lr.01024")
    h = f.halos()
    # hard-code the centre so we're not implicitly testing the centering routine too:
    cen = [0.024456279579533, -0.034112552174141, -0.122436359962132]
    #cen = pynbody.analysis.halo.center(h[1],retcen=True)
    #print "[%.15f, %.15f, %.15f]"%tuple(cen)
    f['pos']-=cen

    # derive smoothing lengths direct from file data so we are
    # not testing the kdtree (which is tested elsewhere)

    f.gas['smooth']=(f.gas['mass']/f.gas['rho'])**(1,3)
    np.save("result_im_x_pre_phys.npy", f.gas['x'])
    f.physical_units()
    np.save("result_im_x_post_phys.npy", f.gas['x'])

@pytest.fixture
def compare2d():
    yield np.load(test_folder / "test_im_2d.npy")

@pytest.fixture
def compare3d():
    yield np.load(test_folder / "test_im_3d.npy")

@pytest.fixture
def compare2d_wendlandC2():
    yield np.load(test_folder / "test_im_2d_wendlandC2.npy")

@pytest.fixture
def compare3d_wendlandC2():
    yield np.load(test_folder / "test_im_3d_wendlandC2.npy")

@pytest.fixture
def compare_grid():
    yield np.load(test_folder / "test_im_grid.npy")

@pytest.fixture
def stars_2d():
    yield np.load(test_folder / "test_stars_2d.npy")

def test_images(compare2d, compare3d, compare_grid, compare2d_wendlandC2, compare3d_wendlandC2):

    global f

    im3d = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-3", noplot=True, approximate_fast=False)

    im2d = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-2", noplot=True, approximate_fast=False)

    im_grid = pynbody.sph.to_3d_grid(f.gas,nx=200,x2=20.0)[::50]


    np.save("result_im_2d.npy",im2d)
    np.save("result_im_3d.npy",im3d)
    np.save("result_im_grid.npy",im_grid)


    npt.assert_allclose(im2d,compare2d,rtol=1e-4)
    npt.assert_allclose(im3d,compare3d,rtol=1e-4)
    npt.assert_allclose(im_grid,compare_grid,rtol=1e-4)

    # Make images with a different kernel (Wendland C2)
    im3d_wendlandC2 = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-3", noplot=True, approximate_fast=False, kernel_type='wendlandC2')
    im2d_wendlandC2 = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-2", noplot=True, approximate_fast=False, kernel_type='wendlandC2')


    np.save("result_im_2d_wendlandC2.npy",im2d_wendlandC2)
    np.save("result_im_3d_wendlandC2.npy",im3d_wendlandC2)

    # Check that using a different kernel produces a different image
    npt.assert_raises(AssertionError,npt.assert_array_equal,im3d_wendlandC2,im3d)
    npt.assert_raises(AssertionError,npt.assert_array_equal,im2d_wendlandC2,im2d)

    # Check that using a different kernel produces the correct image
    npt.assert_allclose(im2d_wendlandC2,compare2d_wendlandC2,rtol=1e-4)
    npt.assert_allclose(im3d_wendlandC2,compare3d_wendlandC2,rtol=1e-4)

    # check rectangular image is OK
    im_rect = pynbody.sph.render_image(f.gas,nx=500,ny=250,x2=10.0,
                                        approximate_fast=False).in_units("m_p cm^-3")
    compare_rect = compare3d[125:-125]
    npt.assert_allclose(im_rect,compare_rect,rtol=1e-4)

def test_approximate_images(compare2d, compare3d):
    global f
    im3d = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-3", noplot=True, approximate_fast=True)
    im2d = pynbody.plot.sph.image(
        f.gas, width=20.0, units="m_p cm^-2", noplot=True, approximate_fast=True )

    np.save("result_approx_im_2d.npy", im2d)
    np.save("result_approx_im_3d.npy", im3d)

    # approximate interpolated images are only close in a mean sense
    assert abs(np.log10(im2d/compare2d)).mean()<0.03
    assert abs(np.log10(im3d/compare3d)).mean()<0.03


def test_denoise_projected_image_throws():
    global f
    # this should be fine:
    pynbody.plot.sph.image(f.gas, width=20.0, units="m_p cm^-3", noplot=True, approximate_fast=True, denoise=True)

    with pytest.raises(ValueError):
        # this should not:
        pynbody.plot.sph.image(f.gas, width=20.0, units="m_p cm^-2", noplot=True, approximate_fast=True, denoise=True)


def test_render_stars(stars_2d):
    global f
    with pytest.warns(UserWarning, match=r"No log file found; reverting to guess-and-check"):
        im = pynbody.plot.stars.render(f, width=10.0, resolution=100, ret_im=True, plot=False)

    np.save("result_stars_2d.npy", im[40:60])

    npt.assert_allclose(stars_2d,im[40:60],atol=0.01)


@pynbody.derived_array
def intentional_circular_reference(sim):
    return sim['intentional_circular_reference']

# Note: we ignore all warnings here, since pytest will otherwise
# trigger a warning internally because of the exception propagation
@pytest.mark.filterwarnings("ignore:.*")
def test_exception_propagation():
    with pytest.raises(RuntimeError):
        pynbody.plot.sph.image(f.gas, qty='intentional_circular_reference')



"""

decomp
======

Tools for bulge/disk/halo decomposition

"""

import logging

import numpy as np

from .. import array, config, filt, util
from . import angmom, profile

logger = logging.getLogger('pynbody.analysis.decomp')


def decomp(h, aligned=False, j_disk_min=0.8, j_disk_max=1.1, E_cut=None, j_circ_from_r=False,
           cen=None, vcen=None, log_interp=False, angmom_size="3 kpc"):
    """
    Creates an array 'decomp' for star particles in the simulation, with an
    integer specifying a kinematic decomposition. The possible values are:

    1 -- thin disk

    2 -- halo

    3 -- bulge

    4 -- thick disk

    5 -- pseudo bulge

    This routine is based on an original IDL procedure by Chris Brook.


    **Parameters:**

    *h* -- the halo to work on

    *j_disk_min* -- the minimum angular momentum as a proportion of
                  the circular angular momentum which a particle must
                  have to be part of the 'disk'

    *j_disk_max* -- the maximum angular momentum as a proportion of
                  the circular angular momentum which a particle can
                  have to be part of the 'disk'

    *E_cut* -- the energy boundary between bulge and spheroid. If
             None, this is taken to be the median energy of the stars.

    *aligned* -- if False, the simulation is recenterd and aligned so
               the disk is in the xy plane as required for the rest of
               the analysis.

    *cen* -- if not None, specifies the center of the halo. Otherwise
           it is found.  This has no effect if aligned=True

    *vcen* -- if not None, specifies the velocity center of the
            halo. Otherwise it is found.  This has no effect if
            aligned=True

    *j_circ_from_r* -- if True, the maximum angular momentum is
    determined as a function of radius, rather than as a function of
    orbital energy. Default False (determine as function of energy).

    *angmom_size* -- the size of the gas sphere used to determine the
     plane of the disk

    """

    import scipy.interpolate as interp
    global config

    # Center, eliminate proper motion, rotate so that
    # gas disk is in X-Y plane
    if not aligned:
        angmom.faceon(h, cen=cen, vcen=vcen, disk_size=angmom_size)

    # Find KE, PE and TE
    ke = h['ke']
    pe = h['phi']

    h['phi'].convert_units(ke.units)  # put PE and TE into same unit system

    te = ke + pe
    h['te'] = te
    te_star = h.star['te']

    te_max = te_star.max()

    # Add an arbitrary offset to the PE to reflect the idea that
    # the group is 'fully bound'.
    te -= te_max
    logger.info("te_max = %.2e" % te_max)

    h['te'] -= te_max

    logger.info("Making disk rotation curve...")

    # Now make a rotation curve for the disk. We'll take everything
    # inside a vertical height of eps*3

    d = h[filt.Disc('1 Mpc', h['eps'].min() * 3)]

    try:

        # attempt to load rotation curve off disk
        r, v_c = np.loadtxt(h.ancestor.filename + ".rot." +
                            str(h.properties["halo_number"]), skiprows=1, unpack=True)

        pro_d = profile.Profile(d, nbins=100, type='log')
        r_me = pro_d["rbins"].in_units("kpc")
        r_x = np.concatenate(([0], r, [r.max() * 2]))
        v_c = np.concatenate(([v_c[0]], v_c, [v_c[-1]]))
        v_c = interp.interp1d(r_x, v_c, bounds_error=False)(r_me)

        logger.info(" - found existing rotation curve on disk, using that")

        v_c = v_c.view(array.SimArray)
        v_c.units = "km s^-1"
        v_c.sim = d

        v_c.convert_units(h['vel'].units)

        pro_d._profiles['v_circ'] = v_c
        pro_d.v_circ_loaded = True

    except Exception:
        pro_d = profile.Profile(d, nbins=100, type='log')  # .D()
        # Nasty hack follows to force the full halo to be used in calculating the
        # gravity (otherwise get incorrect rotation curves)
        pro_d._profiles['v_circ'] = profile.v_circ(pro_d, h)

    pro_phi = pro_d['phi']
    #import pdb; pdb.set_trace()
    # offset the potential as for the te array
    pro_phi -= te_max

    # (will automatically be reflected in E_circ etc)
    # calculating v_circ for j_circ and E_circ is slow

    if j_circ_from_r:
        pro_d.create_particle_array("j_circ", out_sim=h)
        pro_d.create_particle_array("E_circ", out_sim=h)
    else:

        if log_interp:
            j_from_E = interp.interp1d(
                np.log10(-pro_d['E_circ'].in_units(ke.units))[::-1], np.log10(pro_d['j_circ'])[::-1], bounds_error=False)
            h['j_circ'] = 10 ** j_from_E(np.log10(-h['te']))
        else:
            #            j_from_E  = interp.interp1d(-pro_d['E_circ'][::-1], (pro_d['j_circ'])[::-1], bounds_error=False)
            j_from_E = interp.interp1d(
                pro_d['E_circ'].in_units(ke.units), pro_d['j_circ'], bounds_error=False)
            h['j_circ'] = j_from_E(h['te'])

        # The next line forces everything close-to-unbound into the
        # spheroid, as per CB's original script ('get rid of weird
        # outputs', it says).
        h['j_circ'][np.where(h['te'] > pro_d['E_circ'].max())] = np.inf

        # There are only a handful of particles falling into the following
        # category:
        h['j_circ'][np.where(h['te'] < pro_d['E_circ'].min())] = pro_d[
            'j_circ'][0]

    h['jz_by_jzcirc'] = h['j'][:, 2] / h['j_circ']
    h_star = h.star

    if 'decomp' not in h_star:
        h_star._create_array('decomp', dtype=int)
    disk = np.where(
        (h_star['jz_by_jzcirc'] > j_disk_min) * (h_star['jz_by_jzcirc'] < j_disk_max))

    h_star['decomp', disk[0]] = 1
    # h_star = h_star[np.where(h_star['decomp']!=1)]

    # Find disk/spheroid angular momentum cut-off to make spheroid
    # rotational velocity exactly zero.

    V = h_star['vcxy']
    JzJcirc = h_star['jz_by_jzcirc']
    te = h_star['te']

    logger.info("Finding spheroid/disk angular momentum boundary...")

    j_crit = util.bisect(0., 5.0,
                         lambda c: np.mean(V[np.where(JzJcirc < c)]))

    logger.info("j_crit = %.2e" % j_crit)

    if j_crit > j_disk_min:
        logger.warning(
            "!! j_crit exceeds j_disk_min. This is usually a sign that something is going wrong (train-wreck galaxy?)")
        logger.warning("!! j_crit will be reset to j_disk_min=%.2e" % j_disk_min)
        j_crit = j_disk_min

    sphere = np.where(h_star['jz_by_jzcirc'] < j_crit)

    if E_cut is None:
        E_cut = np.median(h_star['te'])

    logger.info("E_cut = %.2e" % E_cut)

    halo = np.where((te > E_cut) * (JzJcirc < j_crit))
    bulge = np.where((te <= E_cut) * (JzJcirc < j_crit))
    pbulge = np.where((te <= E_cut) * (JzJcirc > j_crit)
                      * ((JzJcirc < j_disk_min) + (JzJcirc > j_disk_max)))
    thick = np.where((te > E_cut) * (JzJcirc > j_crit)
                     * ((JzJcirc < j_disk_min) + (JzJcirc > j_disk_max)))

    # h_star['decomp', disk] = 1
    h_star['decomp', halo] = 2
    h_star['decomp', bulge] = 3
    h_star['decomp', thick] = 4
    h_star['decomp', pbulge] = 5

    # Return profile object for informational purposes
    return pro_d



import numpy as np
import numpy.testing as npt
import pytest

import pynbody


@pytest.fixture
def snap():
    f = pynbody.new(1000)
    f['pos'] = np.random.normal(scale=1.0, size=f['pos'].shape)
    f['vel'] = np.random.normal(scale=1.0, size=f['vel'].shape)
    f['mass'] = np.random.uniform(1.0, 10.0, size=f['mass'].shape)
    f['pos'].units = 'kpc'
    f['vel'].units = 'km s^-1'
    f['mass'].units = 'Msol'
    return f


def test_sphere(snap):
    f = snap
    sp = f[pynbody.filt.Sphere(0.5)]
    assert (sp.get_index_list(f) == np.where(f['r'] < 0.5)[0]).all()
    assert sp['x'].max() < 0.5
    assert sp['x'].min() > -0.5
    assert sp['r'].max() < 0.5

    sp_units = f[pynbody.filt.Sphere('500 pc')]
    assert len(sp_units.intersect(sp)) == len(sp)


def test_passfilters(snap):
    f = snap

    hp = f[pynbody.filt.HighPass('mass', 5)]
    lp = f[pynbody.filt.LowPass('mass', 5)]
    bp = f[pynbody.filt.BandPass('mass', 2, 7)]

    assert len(hp) > 0
    assert len(lp) > 0
    assert len(bp) > 0

    assert len(hp.intersect(lp)) == 0

    assert (hp.get_index_list(f) == np.where(f['mass'] > 5)[0]).all()
    assert (lp.get_index_list(f) == np.where(f['mass'] < 5)[0]).all()
    assert (bp.get_index_list(f) == np.where(
        (f['mass'] > 2) * (f['mass'] < 7))[0]).all()

@pytest.mark.parametrize("with_kdtree", [True, False])
def test_cuboid(snap, with_kdtree):
    f = snap
    x1s = np.random.uniform(-0.5, 0.0, 10)
    x2s = np.random.uniform(0.0, 0.5, 10)
    y1s = np.random.uniform(-0.5, 0.0, 10)
    y2s = np.random.uniform(0.0, 0.5, 10)
    z1s = np.random.uniform(-0.5, 0.0, 10)
    z2s = np.random.uniform(0.0, 0.5, 10)

    if with_kdtree:
        f.build_tree()

    for x1, x2, y1, y2, z1, z2 in zip(x1s, x2s, y1s, y2s, z1s, z2s):
        c = f[pynbody.filt.Cuboid(x1, y1, z1, x2, y2, z2)]
        assert (c.get_index_list(f) == np.where(
            (f['x'] > x1) * (f['x'] < x2) * (f['y'] > y1) * (f['y'] < y2) * (f['z'] > z1) * (f['z'] < z2))[0]).all()


def test_logic(snap):
    f = snap

    comp = f[pynbody.filt.BandPass('mass', 2, 7)]

    and_test = f[
        pynbody.filt.HighPass('mass', 2) & pynbody.filt.LowPass('mass', 7)]

    assert and_test == comp

    comp = f[pynbody.filt.LowPass('mass', 2)]
    or_test = f[
        (pynbody.filt.LowPass('mass', 1) | pynbody.filt.BandPass('mass', 1, 2))]

    assert or_test == comp

    comp = f[pynbody.filt.BandPass('mass', 2, 7)]
    not_test = f[~(pynbody.filt.BandPass('mass', 2, 7))]

    assert comp.union(not_test) == f
    assert len(comp.intersect(not_test)) == 0

    assert len(comp)+len(not_test)==len(f)
    assert len(comp)!=0
    assert len(not_test)!=0

def test_family_filter():
    f = pynbody.new(dm=100,gas=100)
    f_dm = f.dm
    f_dm_filter = f[pynbody.filt.FamilyFilter(pynbody.family.dm)]
    f_gas = f.gas
    f_gas_filter = f[pynbody.filt.FamilyFilter(pynbody.family.gas)]
    assert (f_dm.get_index_list(f) == f_dm_filter.get_index_list(f)).all()
    assert (f_gas.get_index_list(f) == f_gas_filter.get_index_list(f)).all()


def test_hashing():
    X = {}
    X[pynbody.filt.Sphere('100 kpc')] = 5

    X[pynbody.filt.FamilyFilter(pynbody.family.gas)] = 10
    assert X.get(pynbody.filt.Sphere('100 kpc'), None) == 5
    assert X.get(pynbody.filt.FamilyFilter(pynbody.family.gas),None)==10
    with pytest.raises(KeyError):
        X[pynbody.filt.FamilyFilter(pynbody.family.dm)]



import time

import numpy as np
import numpy.testing as npt
import pytest

import pynbody


def random_slice(max_pos=1000, max_step=10):
    import random
    a = random.randint(0, max_pos)
    b = random.randint(0, max_pos)
    return slice(min(a, b), max(a, b), random.randint(1, max_step))

def test_thread_map():
    result = pynbody.util.thread_map(lambda x: 2*x, range(5))
    assert (np.array(result) == [0, 2, 4, 6, 8]).all()

def test_thread_map_exception():
    def testfn(x):
        if x==0:
            raise RuntimeError("Test exception")
        return 2*x

    with pytest.raises(RuntimeError) as exc:
        pynbody.util.thread_map(testfn, range(5))

    assert exc.value.args[0] == "Test exception"
    assert exc.traceback[-1].name == 'testfn'

def test_intersect_slices():
    """Unit test for intersect_slices, relative_slice and chained_slice"""
    from pynbody.util import intersect_slices

    numbers = np.arange(1000)

    for x in range(1000):
        s1 = random_slice()
        s2 = random_slice()
        n1 = numbers[s1]
        n2 = numbers[s2]
        s_inter = pynbody.util.intersect_slices(s1, s2, len(numbers))
        nx = numbers[s_inter]
        nx2 = numbers[s1][pynbody.util.relative_slice(s1, s_inter)]
        correct_intersection = set(n1).intersection(set(n2))
        assert set(nx) == correct_intersection
        assert set(nx2) == correct_intersection

        s1 = random_slice()
        s2 = random_slice(20, 4)
        n3 = numbers[s1][s2]
        s3 = pynbody.util.chained_slice(s1, s2)
        nx3 = numbers[s3]
        assert len(n3) == len(nx3)
        assert all([x == y for x, y in zip(n3, nx3)])

        rel_slice_to_self = pynbody.util.relative_slice(s1, s1)
        assert rel_slice_to_self.step is None
        assert rel_slice_to_self.start == 0

    # test 'None' steps handled correctly
    assert pynbody.util.intersect_slices(
        slice(0, 5, None), slice(0, 5, None)) == slice(0, 5, None)


def test_intersect_indices():
    """Unit test for index_before_slice"""

    numbers = np.arange(1000)

    for x in range(1000):
        sl = random_slice()
        t = numbers[sl]
        if len(t) == 0:
            continue
        ind = np.random.randint(0, len(t), 50)
        new_ind = pynbody.util.index_before_slice(sl, ind)

        assert len(numbers[new_ind]) == len(numbers[sl][ind])

        assert all(
            [x == y for x, y in zip(numbers[new_ind], numbers[sl][ind])])


def test_bisect():
    """Unit test for bisection algorithm"""

    assert abs(
        pynbody.util.bisect(0., 1., lambda x: x ** 2 - 0.04) - 0.20) < 1.e-6


def test_slice_length():
    N = np.arange(100)

    for end in range(100):
        for start in range(end):
            for step in range(1, 10):
                S = slice(0, end, step)
                assert len(N[S]) == pynbody.util.indexing_length(S)


def test_IC_grid_gen():
    res1 = pynbody.util.grid_gen(slice(2,100,5),5,5,5)
    correct1 = np.array([[ 0.5,  0.1,  0.1],
       [ 0.5,  0.3,  0.1],
       [ 0.5,  0.5,  0.1],
       [ 0.5,  0.7,  0.1],
       [ 0.5,  0.9,  0.1],
       [ 0.5,  0.1,  0.3],
       [ 0.5,  0.3,  0.3],
       [ 0.5,  0.5,  0.3],
       [ 0.5,  0.7,  0.3],
       [ 0.5,  0.9,  0.3],
       [ 0.5,  0.1,  0.5],
       [ 0.5,  0.3,  0.5],
       [ 0.5,  0.5,  0.5],
       [ 0.5,  0.7,  0.5],
       [ 0.5,  0.9,  0.5],
       [ 0.5,  0.1,  0.7],
       [ 0.5,  0.3,  0.7],
       [ 0.5,  0.5,  0.7],
       [ 0.5,  0.7,  0.7],
       [ 0.5,  0.9,  0.7]])

    npt.assert_almost_equal(res1,correct1)

    res2 = pynbody.util.grid_gen([14,19,22,94],5,5,5)
    correct2 = np.array([[ 0.9,  0.5,  0.1],
       [ 0.9,  0.7,  0.1],
       [ 0.5,  0.9,  0.1],
       [ 0.9,  0.7,  0.7]])

    npt.assert_almost_equal(res2,correct2)


def test_openmp_summations():
    np.random.seed(0)
    a = np.random.normal(size=int(5e7))
    b = np.random.normal(size=int(5e7))

    start = time.time()
    sum_a = np.dot(a,(b>1.0))
    np_timing = time.time()-start

    start = time.time()
    sum_a_ne = pynbody.util.sum_if_gt(a,b,1.0)
    ne_timing = time.time()-start


    print ("NP: %.1f NE: %.1f"%(np_timing*1e3,ne_timing*1e3))
    print ("  vals %.1f %.1f"%(sum_a_ne,sum_a))
    npt.assert_allclose(sum_a_ne,sum_a)

    npt.assert_allclose(pynbody.util.sum_if_lt(a,b,0.2),np.dot(a,b<0.2))
    npt.assert_allclose(pynbody.util.sum(a),np.sum(a))


def test_invert():
    # regression test for failure to invert some matrices

    M = [[1,3,0,0,0],[3,14,-1,0,0],[0,-1,1,0,0],[0,0,0,1,0],[0,0,0,0,1]]
    Minv = pynbody.util.rational_matrix_inv(M)

    assert (np.dot(Minv,M)==np.diag([1]*5)).all()

def test_find_boundaries():
    # regression test for failure of find_boundaries when negative indices are present

    our_numbers = np.array([-1,-1,0,0,0,1,2,2,3,3,5,5,5], dtype=np.int32)
    boundaries = pynbody.util.find_boundaries(our_numbers)
    assert (boundaries==[2,5,6,8,-1,10]).all()

def test_binary_search():
    """Unit test for binary search algorithm"""
    seed = 16091992
    np.random.seed(seed)
    b = np.random.choice(10000, size=10000, replace=False)
    a = np.sort(np.random.choice(b, size=100))

    b_argsort = np.argsort(b)

    # Test that algorithm works with different number of threads
    # for nthreads in range(10, 1, -1):
    for nthreads in range(1, 10):
        indices = pynbody.util.binary_search(a, b, b_argsort, num_threads=nthreads)
        np.testing.assert_array_equal(a, b[indices])

    # Test that algorithm works for missing elements in b
    b = np.random.choice(np.arange(100), size=100, replace=False)
    b_argsort = np.argsort(b)

    a = np.array([0, 10, 100, 1000])
    for nthreads in range(1, 10):
        indices = pynbody.util.binary_search(a, b, b_argsort, num_threads=nthreads)
        mask = (indices != len(b))
        np.testing.assert_array_equal(a[mask], b[indices[mask]])

def test_is_sorted():
    """Unit test for is_sorted function"""
    # Pathological cases
    assert pynbody.util.is_sorted(np.array([])) == 1
    assert pynbody.util.is_sorted(np.array([1])) == 1
    arr = np.zeros(100)
    assert pynbody.util.is_sorted(arr) == 1

    # Constant to begin with, then increasing/decreasing
    arr = np.zeros(100)
    arr[-1] = 1
    assert pynbody.util.is_sorted(arr) == 1
    arr[-1] = -1
    assert pynbody.util.is_sorted(arr) == -1

    # Simple cases
    assert pynbody.util.is_sorted(np.array([1, 2, 3])) == 1
    assert pynbody.util.is_sorted(np.array([1, 2, 1])) == 0
    assert pynbody.util.is_sorted(np.array([3, 2, 1])) == -1



"""

KDTree
======

Provides access to nearest neighbour lists and smoothing lengths.

"""
import logging
import time
import warnings

import numpy as np

from .. import array as ar, config, util
from . import kdmain

logger = logging.getLogger("pynbody.kdtree")

# Boundary type definition must exactly match the C++ definition (see kd.h)
Boundary = np.dtype([
    ('fMin', np.float32, (3,)),
    ('fMax', np.float32, (3,))
])

# KDNode type definition must exactly match the C++ definition (see kd.h)
KDNode = np.dtype([
    ('fSplit', np.float32),
    ('bnd', Boundary),
    ('iDim', np.int32),
    ('pLower', np.intp),
    ('pUpper', np.intp)
])

class KDTree:
    """KDTree can be used for smoothing, interpolating and geometrical queries.

    Most users are unlikely to interact with the KDTree directly, instead using the higher-level
    SPH functionality. To accelerate geometrical queries on a snapshot, try:

    >>> f = pynbody.load('my_snapshot')
    >>> f.build_kdtree()
    >>> f[pynbody.filt.Sphere('10 kpc')] # this will automatically be a faster operation once the tree is built

    Performance statistics can be tested using the performance_kdtree.py script in the tests folder

    Most KDTree operations proceed in parallel, using the number of threads specified in the
    pynbody configuration.
    """

    PROPID_HSM = 1
    PROPID_RHO = 2
    PROPID_QTYMEAN_1D = 3
    PROPID_QTYMEAN_ND = 4
    PROPID_QTYDISP_1D = 5
    PROPID_QTYDISP_ND = 6
    PROPID_QTYDIV  = 7
    PROPID_QTYCURL = 8

    def __init__(self, pos, mass, leafsize=32, boxsize=None, num_threads=None, shared_mem=False):
        """
        Parameters
        ----------
        pos : pynbody.array.SimArray
            Particles positions.
        mass : pynbody.array.SimArray
            Particles mass.
        leafsize : int, optional
            The number of particles in leaf nodes (default 32).
        boxsize : float, optional
            Boxsize (default None)
        num_threads : int, optional
            Number of threads to use when building tree (if None, use configured/detected number of processors).
        shared_mem : bool, optional
            Whether to keep kdtree in shared memory so that it can be shared between processes (default False).
        """

        num_threads = self._set_num_threads(num_threads)

        # get a power of 2 for num_threads to pass to the constructor, because
        # otherwise the workload will not be balanced across threads and they
        # will be wasted
        num_threads_init = 2 ** int(np.log2(num_threads))


        self.leafsize = int(leafsize)
        self.kdtree = kdmain.init(pos, mass, self.leafsize)
        nodes = kdmain.get_node_count(self.kdtree)

        if shared_mem:
            from ..array import shared
            self.kdnodes = shared.make_shared_array(nodes, KDNode)
            self.particle_offsets = shared.make_shared_array(len(pos), np.intp)
        else:
            self.kdnodes = np.empty(nodes, dtype=KDNode)
            self.particle_offsets = np.empty(len(pos), dtype=np.intp)
        kdmain.build(self.kdtree, self.kdnodes, self.particle_offsets, num_threads_init)

        self.boxsize = boxsize
        self._pos = pos

    def _set_num_threads(self, num_threads):
        if num_threads is None:
            num_threads = int(config["number_of_threads"])
        self.num_threads = num_threads
        return num_threads

    def serialize(self):
        return self.leafsize, self.boxsize, self.kdnodes, self.particle_offsets

    @classmethod
    def deserialize(cls, pos, mass, serialized_state, num_threads = None, boxsize=None):
        self = object.__new__(cls)
        self._set_num_threads(num_threads)
        leafsize, boxsize_s, kdnodes, particle_offsets = serialized_state
        if boxsize is not None and abs(boxsize-boxsize_s) > 1e-6:
            warnings.warn("Boxsize in serialized state does not match boxsize passed to deserialize")
        self.kdtree = kdmain.init(pos, mass, leafsize)
        self.leafsize = leafsize
        nodes = kdmain.get_node_count(self.kdtree)
        if len(kdnodes) != nodes:
            raise ValueError("Number of nodes in serialized state does not match number of nodes in kdtree")
        self.kdnodes = kdnodes
        if len(particle_offsets) != len(pos):
            raise ValueError("Number of particle offsets in serialized state does not match number of particles")
        self.particle_offsets = particle_offsets
        self.boxsize = boxsize
        self._pos = pos

        kdmain.import_prebuilt(self.kdtree, self.kdnodes, self.particle_offsets, 0)

        return self



    def particles_in_sphere(self, center, radius):
        """Find particles within a sphere.

        Parameters
        ----------
        center : array_like
            Center of the sphere.
        radius : float
            Radius of the sphere.

        Returns
        -------
        indices : array_like
            Indices of the particles within the sphere.
        """
        smx = kdmain.nn_start(self.kdtree, 1, 1, self.boxsize)

        particle_ids = kdmain.particles_in_sphere(self.kdtree, smx, center[0], center[1], center[2], radius)

        kdmain.nn_stop(self.kdtree, smx)
        return particle_ids

    def nn(self, nn=None):
        """Generator of neighbour list.

        This method provides an interface to the neighbours search of the C-extension kdmain.

        Parameters
        ----------
        nn : int, None
            number of neighbours. If None, default to 64.

        Yields
        ------
        nbr_list : list[int, float, list[int], list[float]]
            Information about the neighbours of a particle.
            List with four elements:
                nbr_list[0] : int
                    the index of the particle in the snapshot's arrays.
                nbr_list[1] : float
                    the smoothing length of the particle (i.e. nbr_list[1] == snap['smooth'][nbr_list[0]]).
                nbr_list[2] : list[int]
                    list of the indexes of the `nn` neighbouring particles.
                nbr_list[3] : list[float]
                    list of distances squared of each neighbouring particles.

            The lists of particle and of the relative distance squared are not sorted.

        """
        if nn is None:
            nn = 64

        smx = kdmain.nn_start(self.kdtree, int(nn), 1, self.boxsize)
        kdmain.domain_decomposition(self.kdtree, 1)

        while True:
            nbr_list = kdmain.nn_next(self.kdtree, smx)
            if nbr_list is None:
                break
            yield nbr_list

        kdmain.nn_stop(self.kdtree, smx)

    def all_nn(self, nn=None):
        """A list of neighbours information for all the particles in the snapshot."""
        return [x for x in self.nn(nn)]

    @staticmethod
    def array_name_to_id(name):
        if name == "smooth":
            return 0
        elif name == "rho":
            return 1
        elif name == "mass":
            return 2
        elif name == "qty":
            return 3
        elif name == "qty_sm":
            return 4
        else:
            raise ValueError("Unknown KDTree array")

    def set_array_ref(self, name, ar):
        if self.array_name_to_id(name) < 3:
            if not np.issubdtype(self._pos.dtype, ar.dtype):
                raise TypeError(
                    "KDTree requires matching dtypes for %s (%s) and pos (%s) arrays"
                    % (name, ar.dtype, self._pos.dtype)
                )
        kdmain.set_arrayref(self.kdtree, self.array_name_to_id(name), ar)
        assert self.get_array_ref(name) is ar

    def get_array_ref(self, name):
        return kdmain.get_arrayref(self.kdtree, self.array_name_to_id(name))

    def smooth_operation_to_id(self, name):
        if name == "hsm":
            return self.PROPID_HSM
        elif name == "rho":
            return self.PROPID_RHO
        elif name == "qty_mean":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) == 1:
                return self.PROPID_QTYMEAN_1D
            elif len(input_array.shape) == 2:
                if input_array.shape[1] != 3:
                    raise ValueError("Currently only able to smooth 3D or 1D arrays")
                return self.PROPID_QTYMEAN_ND
        elif name == "qty_disp":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) == 1:
                return self.PROPID_QTYDISP_1D
            elif len(input_array.shape) == 2:
                if input_array.shape[1] != 3:
                    raise ValueError("Currently only able to smooth 3D or 1D arrays")
                return self.PROPID_QTYDISP_ND
        elif name == "qty_div":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) != 2 and input_array.shape[1] != 3:
                raise ValueError("Can only compute divergence of 3D arrays")
            return self.PROPID_QTYDIV
        elif name == "qty_curl":
            input_array = self.get_array_ref("qty")
            if len(input_array.shape) != 2 and input_array.shape[1] != 3:
                raise ValueError("Can only compute curl of 3D arrays")
            return self.PROPID_QTYCURL
        else:
            raise ValueError("Unknown smoothing request %s" % name)

    def populate(self, mode, nn, kernel = 'CubicSpline'):
        """Create the KDTree and perform the operation specified by `mode`.

        Parameters
        --------
        mode : str (see `kdtree.smooth_operation_to_id`)
            Specify operation to perform (compute smoothing lengths, density, SPH mean, or SPH dispersion).
        nn : int
            Number of neighbours to be considered when smoothing.
        kernel : str
            Keyword to specify the smoothing kernel. Options: 'CubicSpline', 'WendlandC2'
        """


        if nn is None:
            nn = 64

        smx = kdmain.nn_start(self.kdtree, int(nn), self.num_threads, self.boxsize)

        try:
            propid = self.smooth_operation_to_id(mode)

            if propid == self.PROPID_HSM:
                kdmain.domain_decomposition(self.kdtree, self.num_threads)

            if kernel == 'CubicSpline':
                kernel = 0
            elif kernel == 'WendlandC2':
                kernel = 1
            else:
                raise ValueError(
                    "Kernel keyword %s not recognised. Please choose either 'CubicSpline' or 'WendlandC2'." % kernel
                )

            if self.num_threads == 1:
                kdmain.populate(self.kdtree, smx, propid, 0, kernel)
            else:
                util.thread_map(
                    kdmain.populate,
                    [self.kdtree] * self.num_threads,
                    [smx] * self.num_threads,
                    [propid] * self.num_threads,
                    list(range(0, self.num_threads)),
                    [kernel] * self.num_threads
                )
        finally:
            # Free C-structures memory
            kdmain.nn_stop(self.kdtree, smx)

    def sph_mean(self, array, nsmooth=64, kernel = 'CubicSpline'):
        r"""Calculate the SPH mean of a simulation array.

        It's the application of the SPH interpolation formula for computing the smoothed quantity at particles position.
        It uses the cubic spline smoothing kernel W.

            qty_sm[i] = \sum_{j=0}^{nsmooth}( mass[j]/rho[j] * qty[i] * W(|pos[i] - pos[j]|, smooth[i]) )

        Actual computation is done in the C-extension functions smooth.cpp::smMeanQty[1,N]D.

        Parameters
        ----------
        array : pynbody.array.SimArray
            Quantity to smooth (compute SPH interpolation at particles position).
        nsmooth:
            Number of neighbours to use when smoothing.

        Returns
        -------
        output : pynbody.array.SimArray
            The SPH mean of the input array.
        """
        output = np.empty_like(array)

        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Smoothing array with %d nearest neighbours" % nsmooth)
        start = time.time()
        self.populate("qty_mean", nsmooth, kernel)
        end = time.time()

        logger.info("SPH smooth done in %5.3g s" % (end - start))

        return output

    def sph_dispersion(self, array, nsmooth=64, kernel = 'CubicSpline'):
        r"""Calculate the SPH dispersion of a simulation array.

        It uses the cubic spline smoothing kernel W.

        First it computes the smoothed quantity:

            qty_sm[i] = \sum_{j=0}^{nsmooth}( mass[j]/rho[j] * qty[i] * W(|pos[i] - pos[j]|, smooth[i]) )

        Then the squared root of the SPH smoothed variance:

            qty_disp[i] = \sqrt( \sum_{j=0}^{nsmooth}( mass[j]/rho[j] * (qty_sm[i] - qty[j])^2 * W(|pos[i] - pos[j]|, smooth[i]) ) )

        Actual computation is done in the C-extension functions smooth.cpp::smDispQty[1,N]D.

        Parameters
        ----------
        array : pynbody.array.SimArray
            Quantity to compute dispersion of.
        nsmooth:
            Number of neighbours to use when smoothing.

        Returns
        -------
        output : pynbody.array.SimArray
            The dispersion of the input array.
        """
        output = np.empty_like(array)
        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Getting dispersion of array with %d nearest neighbours" % nsmooth)
        start = time.time()
        self.populate("qty_disp", nsmooth, kernel)
        end = time.time()

        logger.info("SPH dispersion done in %5.3g s" % (end - start))

        return output

    def _sph_differential_operator(self, array, op, nsmooth=64, kernel = 'CubicSpline'):
        if op == "div":
            op_label = "divergence"
            output = np.empty(len(array), dtype=array.dtype)
        elif op == "curl":
            op_label = "curl"
            output = np.empty_like(array)
        else:
            raise ValueError("Can only compute curl or divergence")

        if hasattr(array, "units"):
            output = output.view(ar.SimArray)
            output.units = array.units/self._pos.units

        self.set_array_ref("qty", array)
        self.set_array_ref("qty_sm", output)

        logger.info("Getting %s of array with %d nearest neighbours" % (op_label, nsmooth))
        start = time.time()
        self.populate("qty_%s" % op, nsmooth, kernel)
        end = time.time()

        logger.info(f"SPH {op_label} done in {end - start:5.3g} s")

        return output

    def sph_curl(self, array, nsmooth=64, kernel = 'CubicSpline'):
        return self._sph_differential_operator(array, "curl", nsmooth, kernel)

    def sph_divergence(self, array, nsmooth=64, kernel = 'CubicSpline'):
        return self._sph_differential_operator(array, "div", nsmooth, kernel)

    def __del__(self):
        if hasattr(self, "kdtree"):
            kdmain.free(self.kdtree)



import glob
import pickle
import sys

import matplotlib.pyplot as plt
import numpy as np

import pynbody

runs = ['c.1td.05rp.1','mugs','esn1.2','td.01c.1rp.1','rp.175','noradpres','2crasy']
labels = ['Fiducial','MUGS','120% SN energy','Low Diffusion','High Diffusion','No ESF',
          'High ESF']
cs = ['r','k','c','g','y','m','b']
for j,r in enumerate(runs):
    simname=r+'/01024/g1536.01024'

    d = pickle.load(open(simname+'.data'))

    if j>0:
        plt.plot(d['rotcur']['r'],d['rotcur']['vc'],color=cs[j],
                 lw=1, label=labels[j],alpha=0.5)
        if j==6:
            rad = d['rotcur']['r']
            plt.plot(rad,(2.0/np.pi)*np.max(d['rotcur']['vc'])*np.arctan(rad),
                     '--b',label='arctangent fit')
    else:
        plt.plot(d['rotcur']['r'],d['rotcur']['vc'],color=cs[j],
                 lw=2, label=labels[j])
        rad = d['rotcur']['r']
        plt.plot(rad,(2.0/np.pi)*np.max(d['rotcur']['vc'])*np.arctan(4*rad),
                 '--r',label='arctangent fit')

plt.savefig('rc.eps')



"""

metals
======

"""


import numpy as np

from .generic import gauss_kde


def mdf(sim, filename=None, clear=True, range=[-5, 0.3], axes=False, **kwargs):
    '''

    Metallicity Distribution Function

    Plots a metallicity distribution function to the best of
    matplotlib's abilities.  Unfortunately, the "normed" keyword is
    buggy and does not return a PDF.  The "density" keyword should,
    but it not yet supported in many versions of numpy.

    **Usage:**

    >>> import pynbody.plot as pp
    >>> pp.mdf(s,linestyle='dashed',color='k')


    '''
    nbins = 100
    if axes:
        plt = axes
    else:
        import matplotlib.pyplot as plt
        if clear:
            plt.clf()
        plt.xlabel('[Fe / H]')
        plt.ylabel('PDF')

    metpdf, bins = np.histogram(sim['feh'], weights=sim['mass'],
                                bins=nbins, normed=True, range=range, **kwargs)  # density=True,
    midpoints = 0.5 * (bins[:-1] + bins[1:])

    plt.plot(midpoints, metpdf)
    if (filename):
        print("Saving " + filename)
        plt.savefig(filename)


def ofefeh(sim, fxn=gauss_kde, filename=None, **kwargs):
    '''

    Use :func:`~pynbody.plot.generic.hist2d` to make a [O/Fe] vs. [Fe/H] plot

    **Input:**

    *sim*: snapshot to pull data from

    **Optional Keywords:**

    *fxn*: a function with the same signature as functions
       :func:`~pynbody.plot.generic.hist2d` and
       :func:`~pynbody.plot.generic.gauss_kde` in
       :mod:`pynbody.plot.generic` default:
       :func:`pynbody.plot.generic.hist2d`


    see :func:`~pynbody.plot.generic.make_contour_plot` for other
    plot-related keywords.

    '''

    if 'subplot' in kwargs:
        fxn(sim['feh'], sim['ofe'], filename=filename, **kwargs)
    else:
        fxn(sim['feh'], sim['ofe'], filename=filename,
            xlabel="[Fe/H]", ylabel="[O/Fe]", **kwargs)
